# -*- coding: utf-8 -*-
"""
This module reads and process data from 24-bit Biosemi BDF files recorded with the ActiveTwo system together with
provided stimulation info.
@author:Jaime Undurraga
"""

from peegy.io.readers import edf_bdf_reader as eb_reader
from peegy.layouts import layouts as lay
from peegy.definitions.eegReaderAbstractClasses import EEGData, EegChannel
import numpy as np
from peegy.processing.tools.filters.resampling import eeg_resampling
from peegy.plot.eeg_ave_epochs_plot_tools import eeg_save_time_slice
from prettytable import PrettyTable
import logging
import astropy as u
from numpy import empty as _empty, float32
import os
import matplotlib.pyplot as plt

log = logging.getLogger()


def empty(*args, **kwargs):
    kwargs.update(dtype=float32)
    _empty(*args, **kwargs)


class EdfBdfDataReader(EEGData):
    def __init__(self, **kwargs):
        super(EdfBdfDataReader, self).__init__(**kwargs)
        self.layout_setup = kwargs.get('layout_setup', 'biosemi64.lay')
        self.channels_to_process = kwargs.get('channels_to_process', [])
        layout_replacement_list = kwargs.get('layout_replacement_list', None)
        self._data = []
        self._channelLabels = []
        _, file_extension = os.path.splitext(self.file_name)
        if file_extension == '.bdf':
            self.type = 'bdf'
        else:
            self.type = 'edf'
        self._header = eb_reader.read_edf_bdf_header(self.file_name)
        self.n_Channels = self._header['n_channels']
        self._fs = self._header['fs'][0]
        self._fs_raw = self._header['fs'][0]
        self._time_units = u.s
        self._amplitude_units = 0.0001 * u.V
        self._channels = []
        self._bad_channels = []
        self._oeg_channels = []
        self.layout = lay.Layout(file_name=self.layout_setup)

        self.data_processing['labels_to_replace'] = np.ndarray.tolist(layout_replacement_list) if \
            layout_replacement_list is not None else None

        if layout_replacement_list is not None:
            lay.change_layout_electrode(self.layout,
                                        labels_to_replace=layout_replacement_list,
                                        channel_mapping=self._header['channels'])

        self._channels = np.array([EegChannel(number=_ch) for _ch in self.channels_to_process])
        for _ch in self._channels:
            if _ch.number in self.layout.get_index():
                idx = int(np.argwhere(self.layout.get_index() == _ch.number))
                _item = self.layout.get_item(idx)
                _ch.x, _ch.y, _ch.h, _ch.label = (_item.x, _item.h, _item.y, _item.label)
            else:
                ch_nums = [_biosemi_ch['number'] for _biosemi_ch in self._header['channels']]
                idx = ch_nums.index(_ch.number)
                _ch.label = self._header['channels'][idx]['label']
                _ch.number = self._header['channels'][idx]['number']

        if len(self.ref_channels) == 1 and self.ref_channels[0].number is None:
            self.ref_channels = self._channels

    def get_all_channels(self):
        # TODO filter bad channels and oeg channels
        return self._channels

    all_channels = property(get_all_channels)

    def get_channel_by_label(self, channel_labels=['']):
        _channels = np.array([EegChannel(label=_ch) for _ch in channel_labels])
        for _ch in _channels:
            if _ch.label in self.layout['label']:
                idx = self.layout['label'].index(_ch.label)
                _ch.x, _ch.y, _ch.h, _ch.number = (self.layout['x'][idx], self.layout['y'][idx],
                                                   self.layout['h'][idx], self.layout['channel'][idx])
            else:
                ch_labels = [_biosemi_ch['label'] for _biosemi_ch in self._header['channels']]
                idx = ch_labels.index(_ch.label)
                _ch.number = self._header['channels'][idx]['number']
        return _channels

    def get_channel_idx(self, channel_labels=None):
        channels = [_ch for _ch in self._header['channels'] if _ch['label'] in channel_labels]
        return channels

    def get_triggers_events(self, **kwargs):
        trigger_channels = kwargs.get('trigger_channels', None)
        ideal_epoch_length = kwargs.get('ideal_epoch_length', None)
        if trigger_channels is not None:
            _channels = self.get_channel_idx(channel_labels=trigger_channels)
            trigger_data = eb_reader.get_data(header=self._header, channels=_channels)
            fig_file_name = self.paths.figure_basename_path + 'detected_triggers' + '.png'
            _events = self.find_triggers(data=trigger_data[:, 0],
                                         ideal_epoch_length=ideal_epoch_length,
                                         fig_file_name=fig_file_name)

        else:
            _, _events, _ = eb_reader.get_trigger_channel(header=self._header)

        # if requested we append new triggers
        _events = self.append_trigger_events(_events, **kwargs)
        return _events

    def append_trigger_events(self, events, **kwargs):
        target_event = kwargs.get('target_event', None)
        new_event_code = kwargs.get('new_event_code', None)
        new_event_time = kwargs.get('new_event_time', None)
        if target_event is None or new_event_code is None:
            return events
        code = np.array([])
        idx = np.array([])
        dur = np.array([])
        dur_samples = np.array([])
        for _code, _idx, _dur, _dur_samples in zip(events['code'], events['idx'], events['dur'], events['dur_samples']):
            code = np.append(code, _code)
            idx = np.append(idx, _idx)
            dur = np.append(dur, _dur)
            dur_samples = np.append(dur_samples, _dur_samples)
            if _code == target_event:
                code = np.append(code, new_event_code)
                idx = np.append(idx, np.round(new_event_time * self.fs).astype(int) + _idx)
                dur = np.append(dur, _dur)
                dur_samples = np.append(dur_samples, _dur_samples)

        event_table = {'code': code,

                       'idx': idx,
                       'dur': dur,
                       'dur_samples': dur_samples}
        return event_table

    def get_triggers(self, trigger_codes=[], **kwargs):
        _trigger_data = eb_reader.get_trigger_channel(header=self._header, **kwargs)
        return self.get_sub_triggers(trigger_data=_trigger_data, trigger_codes=trigger_codes)

    def get_sub_triggers(self, trigger_data=[], trigger_codes=[]):
        if not trigger_codes:
            sub_set_idx = np.arange(len(trigger_data['code'])).astype(int)
        else:
            sub_set_idx = np.where(trigger_data['code'] == trigger_codes[0])[0]
            if not sub_set_idx.size or not sub_set_idx.any():
                _triggers = {'idx': np.array([]),
                             'dur': np.array([]),
                             'code': np.array([]),
                             'dur_samples': np.array([]),
                             'min_distance': np.array([]),
                             'max_distance': np.array([]),
                             'position': np.array([]),
                             'triggers': []}
                return _triggers
        if len(trigger_data['idx'][sub_set_idx]) == 1:
            if len(trigger_data['idx']) > sub_set_idx + 1:
                min_distance = trigger_data['idx'][sub_set_idx + 1] - trigger_data['idx'][sub_set_idx]
                max_distance = min_distance
            elif sub_set_idx - 1 >= 0:
                min_distance = trigger_data['idx'][sub_set_idx] - trigger_data['idx'][sub_set_idx - 1]
                max_distance = min_distance
        else:
            min_distance = min(np.diff(trigger_data['idx'][sub_set_idx]))
            max_distance = max(np.diff(trigger_data['idx'][sub_set_idx]))

        _triggers = {'idx': trigger_data['idx'][sub_set_idx],
                     'dur': trigger_data['dur'][sub_set_idx],
                     'code': trigger_data['code'][sub_set_idx],
                     'dur_samples': trigger_data['dur_samples'][sub_set_idx],
                     'min_distance': [min_distance] * len(trigger_data['idx'][sub_set_idx]),
                     'max_distance': [max_distance] * len(trigger_data['idx'][sub_set_idx]),
                     'position': np.array(sub_set_idx),
                     'triggers': []}

        if len(trigger_codes[1:]) != 0:
            for i in range(len(_triggers['idx'])):
                ini_idx = _triggers['position'][i]
                if i == len(_triggers['idx']) - 1:
                    end_idx = -1
                else:
                    end_idx = _triggers['position'][i + 1] + 1
                _sub_triggers = {}
                for j, key_name in enumerate(trigger_data.keys()):
                    _sub_triggers[key_name] = trigger_data[key_name][ini_idx:end_idx]

                new_triggers = self.get_sub_triggers(trigger_data=_sub_triggers, trigger_codes=trigger_codes[1:])
                _triggers['triggers'].append(new_triggers)
        return _triggers

    @staticmethod
    def get_all_triggers(trigger_data=[], trigger_codes=[]):
        _triggers = {'idx': np.array([]),
                     'dur': np.array([]),
                     'code': np.array([]),
                     'dur_samples': np.array([]),
                     'min_distance': np.array([]),
                     'max_distance': np.array([]),
                     'position': np.array([]),
                     'triggers': []}
        for _code in trigger_codes:
            sub_set_idx = np.where(trigger_data['code'] == _code)[0]
            if not sub_set_idx.size or not sub_set_idx.any():
                continue
            if len(trigger_data['idx'][sub_set_idx]) == 1:
                if len(trigger_data['idx']) > sub_set_idx + 1:
                    min_distance = trigger_data['idx'][sub_set_idx + 1] - trigger_data['idx'][sub_set_idx]
                    max_distance = min_distance
                elif sub_set_idx - 1 >= 0:
                    min_distance = trigger_data['idx'][sub_set_idx] - trigger_data['idx'][sub_set_idx - 1]
                    max_distance = min_distance
            else:
                min_distance = min(np.diff(trigger_data['idx'][sub_set_idx]))
                max_distance = max(np.diff(trigger_data['idx'][sub_set_idx]))

            f_triggers = {'idx': trigger_data['idx'][sub_set_idx],
                          'dur': trigger_data['dur'][sub_set_idx],
                          'code': trigger_data['code'][sub_set_idx],
                          'dur_samples': trigger_data['dur_samples'][sub_set_idx],
                          'min_distance': [min_distance] * len(trigger_data['idx'][sub_set_idx]),
                          'max_distance': [max_distance] * len(trigger_data['idx'][sub_set_idx]),
                          'position': np.array(sub_set_idx),
                          'triggers': []}
            _triggers['triggers'].append(f_triggers)
        return _triggers

    def merge_triggers(self, triggers, trigger_code=[], new_code=None, tolerance=1):
        _new_triggers = []
        new_code = trigger_code if new_code is None else new_code

        for _idx, _code in enumerate(triggers['code'].astype(int)):
            if _code & trigger_code == trigger_code:
                if len(_new_triggers) == 0:
                    _new_triggers = {'idx': np.array([triggers['idx'][_idx]]),
                                     'dur': np.array([triggers['dur'][_idx]]),
                                     'code': np.array([new_code]),
                                     'dur_samples': np.array([triggers['dur_samples'][_idx]]),
                                     'min_distance': -1,
                                     'max_distance': -1,
                                     'position': np.array([triggers['position'][_idx]]),
                                     'triggers': []}
                elif triggers['idx'][_idx] - (_new_triggers['idx'][-1] + _new_triggers['dur_samples'][-1]) > tolerance:
                    _new_triggers['idx'] = np.append(_new_triggers['idx'], triggers['idx'][_idx])
                    _new_triggers['dur'] = np.append(_new_triggers['dur'], triggers['dur'][_idx])
                    _new_triggers['dur_samples'] = np.append(_new_triggers['dur_samples'],
                                                             triggers['dur_samples'][_idx])
                    _new_triggers['position'] = np.append(_new_triggers['position'], triggers['position'][_idx])
                    _new_triggers['code'] = np.append(_new_triggers['code'], new_code)
                else:
                    _new_triggers['dur'][-1] += triggers['dur'][_idx]
                    _new_triggers['dur_samples'][-1] += triggers['dur_samples'][_idx]
        min_distance = min(np.diff(_new_triggers['idx']))
        max_distance = max(np.diff(_new_triggers['idx']))
        _new_triggers['min_distance'] = min_distance
        _new_triggers['max_distance'] = max_distance

        new_unique_events = np.unique(_new_triggers['code'])
        new_event_counter = []
        for j, code in enumerate(new_unique_events):
            new_event_counter.append({'code': code, 'n': len(np.where(triggers['code'] == code)[0])})
        logging.info("\n".join([self.file_name, 'Merged Trigger events:', str(new_event_counter)]))
        return _new_triggers

    def split_triggers_by_std(self, triggers, epoch_code=None, std_thr=1.0):
        unique_events = np.unique(triggers['code'])
        # rank triggers
        _triggers_pos = triggers['idx'][triggers['code'] == epoch_code]
        _triggers_diff = np.diff(_triggers_pos)

        _positions_above_thr = np.where(_triggers_diff > std_thr * np.std(_triggers_diff) + np.mean(_triggers_diff))[0]

        if _positions_above_thr.size:
            _positions_above_thr += 1
            target_pos = np.unique(_triggers_pos[_positions_above_thr])
            # set new trigger code for found triggers
            new_code = unique_events[-1] + 1
            triggers['code'][0] = new_code
            # create an event before detected trigger sections
            for _t_pos in target_pos:
                _match_events = int(np.where(triggers['idx'] == _t_pos)[0])
                triggers = self.insert_trigger_event(triggers=triggers, idx=_match_events, code=new_code)

            new_triggers = self.get_sub_triggers(trigger_data=triggers, trigger_codes=[new_code, epoch_code])
            _message = 'there were %i triggers which deviated from the mean trigger duration of %f samples' % \
                       (len(target_pos), np.mean(_triggers_diff))

            logging.info(_message)
            print(_message)
        else:
            new_triggers = self.get_sub_triggers(trigger_data=triggers, trigger_codes=[epoch_code])

        event_counter = []
        for j, code in enumerate(unique_events):
            event_counter.append({'code': code, 'n': len(np.where(new_triggers['code'] == code)[0])})
        logging.info("\n".join([self.file_name, 'Trigger events:', str(event_counter)]))

        return new_triggers

    @staticmethod
    def insert_trigger_event(triggers={}, idx=None, code=None):
        # insert event before given idx
        aux_triggers = {}
        for _key in list(triggers.keys()):
            aux_triggers[_key] = triggers[_key][0:idx]
        aux_triggers['dur'] = np.append(aux_triggers['dur'], triggers['dur'][idx])
        aux_triggers['code'] = np.append(aux_triggers['code'], code)
        aux_triggers['idx'] = np.append(aux_triggers['idx'], triggers['idx'][idx] - triggers['dur_samples'][idx] - 1)
        aux_triggers['dur_samples'] = np.append(aux_triggers['dur_samples'], triggers['dur_samples'][idx])
        for _key in list(triggers.keys()):
            aux_triggers[_key] = np.concatenate((aux_triggers[_key], triggers[_key][idx:]))
        return aux_triggers

    def add_trigger_events(self,
                           triggers={},
                           master_triggers=np.array([]),
                           relative_time_positions=np.array([]),
                           new_code_events=np.array([])):
        # add triggers as desired
        assert new_code_events.ndim == 2, "new_code_events must be a 2D array"
        assert new_code_events.shape[0] == len(master_triggers), \
            "number of new_code_events rows must be same as number of master_triggers"
        assert new_code_events.shape[1] == relative_time_positions.size, \
            "The new_code event must have same length as relative_time_positions"
        _idx_rel_samples = self.time_to_samples(relative_time_positions)
        c_triggers = triggers
        for _mt_idx, m_t in enumerate(master_triggers):
            aux_triggers = {'code': [], 'idx': [], 'dur': [], 'dur_samples': []}
            for _i, (_code, _idx, _dur, _dur_samples) in enumerate(zip(c_triggers['code'],
                                                                       c_triggers['idx'],
                                                                       c_triggers['dur'],
                                                                       c_triggers['dur_samples'])):
                aux_triggers['code'] = np.append(aux_triggers['code'], _code)
                aux_triggers['idx'] = np.append(aux_triggers['idx'], _idx)
                aux_triggers['dur'] = np.append(aux_triggers['dur'], _dur)
                aux_triggers['dur_samples'] = np.append(aux_triggers['dur_samples'], _dur_samples)
                if _code == m_t:
                    aux_triggers['code'] = np.append(aux_triggers['code'], new_code_events[_mt_idx])
                    aux_triggers['idx'] = np.append(aux_triggers['idx'], _idx + _idx_rel_samples)
                    aux_triggers['dur'] = np.append(aux_triggers['dur'], _dur * np.ones(_idx_rel_samples.shape))
                    aux_triggers['dur_samples'] = np.append(aux_triggers['dur_samples'],
                                                            _dur_samples * np.ones(_idx_rel_samples.shape))
            c_triggers = aux_triggers
        new_epoch_codes = np.concatenate(new_code_events)
        return c_triggers, new_epoch_codes

    @staticmethod
    def clean_triggers(triggers={}, **kwargs):
        min_number_triggers = kwargs.get('min_number_triggers', 10)
        sub_n_triggers = [len(_sub_trigg['idx']) for _, _sub_trigg in enumerate(triggers['triggers'])]
        sub_codes = [np.unique(_sub_trigg['code']) for _, _sub_trigg in enumerate(triggers['triggers'])]
        to_remove = np.where(np.array(sub_n_triggers) < min_number_triggers)[0]
        t = PrettyTable()
        if triggers['code'].size:
            t.add_column(fieldname='Parent Trigger code', column=triggers['code'])
        if sub_codes:
            t.add_column(fieldname='Sub-trigger code', column=sub_codes)
        if sub_n_triggers:
            t.add_column(fieldname='Number of sub-triggers', column=sub_n_triggers)
            t.add_column(fieldname='Used', column=np.array(sub_n_triggers) >= min_number_triggers)
        print(t)
        logging.info(t)
        for _i, _idx_rem in enumerate(to_remove):
            triggers_to_remove = int(_idx_rem - _i)
            print('Removing trigger event %i with %i sub events,   check your data!' % (
                triggers['code'][triggers_to_remove], len(triggers['triggers'][triggers_to_remove]['idx'])))
            for _, _key in enumerate(triggers.keys()):
                triggers[_key] = np.delete(triggers[_key], triggers_to_remove)
        return triggers

    def get_data(self, channels=[{}], triggers={}, **kwargs):
        # auto_detect_reference = kwargs.get('auto_detect_reference', False)
        ini_time = kwargs.get('ini_time', np.floor(triggers['idx'][0] / self._fs_raw))
        end_time = kwargs.get('end_time', np.ceil(
            (triggers['idx'][-1] + triggers['min_distance'][0] - 1) / self._fs_raw))

        data = eb_reader.get_data(header=self._header,
                                  channels=[_ch.__dict__ for _ch in channels],
                                  ini_time=ini_time,
                                  end_time=end_time)
        relative_offset = np.floor(ini_time * self._fs_raw).astype(int)
        _channels_str = '_to_'.join(map(str, [_ch.label for _ch in [channels[0], channels[-1]]]))
        eeg_save_time_slice(data=data,
                            fs=self._fs_raw,
                            channels=self.channels,
                            title='raw data slice',
                            file_name=self.paths.figure_basename_path + _channels_str + 'raw_slice.png')

        t = PrettyTable()
        t.add_column(fieldname='File name', column=[self._file_name])
        t.add_column(fieldname='Trigger block', column=[str(triggers['idx'][0]) + ',' + str(triggers['idx'][-1])])
        log.info(t)
        print(t)
        return data, relative_offset

    def get_raw_data(self, channels=np.array([EegChannel()]), triggers={}, **kwargs):
        ref_channels = kwargs.get('ref_channels', np.array([EegChannel()]))
        resampling_fs = kwargs.get('resampling_fs', None)
        bad_channels = kwargs.get('bad_channels', [])
        remove_bad_channels = kwargs.get('remove_bad_channels', True)
        interpolation_rate = kwargs.get('interpolation_rate', None)
        interpolation_width = kwargs.get('interpolation_width', None)
        interpolation_offset = kwargs.get('interpolation_offset', 0)
        # add reference channels if buffer is not filled and if not included in all channels
        if not self._reference.size:
            for _r_ch in ref_channels:
                if not [_ch for _ch in channels if _ch.__dict__ == _r_ch.__dict__]:
                    channels.append(_r_ch)

        # remove from list bad known channels
        if remove_bad_channels:
            _known_bad_channels_pos = [i for i, _ch in enumerate(channels) if _ch.label in bad_channels]
            _bad_channels = [channels[i] for i in _known_bad_channels_pos]
            if _bad_channels:
                [channels.pop(idx) for bd in _bad_channels for idx, ch in enumerate(channels) if ch == bd]

        if resampling_fs:
            self.resampling_factor = resampling_fs / self._fs_raw

        _data = self.find_array(parameters=self.data_processing, descriptor='raw_data')
        if _data:
            data = _data['raw_data']
            relative_triggers = _data['relative_triggers']
            buffer_size = _data['buffer_size']
            self._fs = self._fs_raw * self.resampling_factor
        else:
            # get data slice
            data, relative_triggers,  buffer_size = \
                self.get_data_slice(channels=channels, triggers=triggers, **kwargs)

            if interpolation_rate is not None and interpolation_width is not None:
                _ini_points = np.array([], dtype=int)
                # _length = np.mean(np.diff(relative_triggers))
                # int_vector = np.floor(np.arange(0, _length / self._fs_raw, 1. / interpolation_rate) * self._fs_raw)
                for _idx, _int_dur in enumerate(np.diff(relative_triggers)):
                    int_vector = np.floor(np.arange(0, _int_dur / self._fs_raw, 1. / interpolation_rate) * self._fs_raw)
                    _ini_points = np.concatenate((_ini_points, relative_triggers[_idx] + int_vector)).astype(int)

                # after last trigger  data is interpolated as long as the mean interval duration
                int_vector = np.floor(
                    np.arange(0, np.mean(np.diff(relative_triggers)) / self._fs_raw, 1. / interpolation_rate) *
                    self._fs_raw)
                _ini_points = np.concatenate((_ini_points, relative_triggers[-1] + int_vector)).astype(int)

                _end_points = _ini_points + np.ceil(interpolation_width * self._fs_raw).astype(int)
                offset = int(np.round(interpolation_offset * self._fs_raw))
                interpolation_data_points = {'ini': _ini_points + offset, 'end': _end_points + offset}
                self.interpolate_data(data, interpolation_data_points=interpolation_data_points)

            # resampling data
            data, _factor = eeg_resampling(x=data, new_fs=resampling_fs, fs=self._fs_raw)
            if _factor != 1.0:
                self.resampling_factor = _factor
                self._fs = self._fs_raw * _factor
                buffer_size = int(buffer_size * _factor)
                relative_triggers = (relative_triggers * _factor).astype(int)

            self.save_array(parameters=self.data_processing, data_type='array',
                            data={'raw_data': data,
                                  'relative_triggers': relative_triggers,
                                  'buffer_size': np.array([buffer_size])},
                            descriptor='raw_data')

        t = PrettyTable()
        t.add_column(fieldname='File name', column=[self._file_name])
        t.add_column(fieldname='Trigger block', column=[str(triggers['idx'][0]) + ',' + str(triggers['idx'][-1])])
        t.add_column(fieldname='Removed channels', column=[bad_channels])
        t.add_column(fieldname='Resampling fs', column=[resampling_fs])
        log.info(t)
        return data, relative_triggers, buffer_size

    def get_data_slice(self, channels=[{}], triggers={}, epoch_length=None, **kwargs):
        fft_size = kwargs.get('fft_size', [])
        if not epoch_length:
            epoch_length = triggers['min_distance'][0] / self._fs_raw
        # epoch_length = np.minimum(epoch_length, triggers['min_distance'][0] / self._fs_raw)
        buffer_size = np.floor(epoch_length * self._fs_raw).astype(int)
        trigger_step = 1.0
        if fft_size:
            trigger_step = np.round(float(fft_size) / float(buffer_size))
            buffer_size = np.floor(epoch_length * trigger_step * self._fs_raw).astype(int)
        # check that trigger + buffer_size fit in the whole recording
        # we check against 2 times buffer_size to leave extra gap at the end
        used_trigger = triggers['idx'][self._header['duration'] * self._fs_raw -
                                       (triggers['idx'] + 2 * buffer_size) >= 0]
        used_trigger = used_trigger[np.arange(0, len(used_trigger), trigger_step).astype(int)]
        # get data. we read up to last trigger plus 2 times buffer_size to leave extra gap at the end
        data, relative_offset = self.get_data(channels=channels, triggers=triggers,
                                              end_time=np.ceil((used_trigger[-1] + 2 * buffer_size - 1) / self._fs_raw),
                                              **kwargs)
        t = PrettyTable()
        t.add_column(fieldname='File name', column=[self._file_name])
        t.add_column(fieldname='Trigger block', column=[str(used_trigger[0]) + ',' + str(used_trigger[-1])])
        t.add_column(fieldname='Number triggers', column=[len(used_trigger)])
        t.add_column(fieldname='Epoch length', column=[epoch_length * trigger_step])
        log.info(t)
        print(t)
        relative_triggers = used_trigger - relative_offset
        return data, relative_triggers, buffer_size

    def set_data(self, new_value):
        self._data = new_value

    data = property(get_data, set_data)

    def get_channel_labels(self):
        return [channel.label for channel in self._channels]

    def set_channel_labels(self, number, new_value):
        if len(number) == len(new_value):
            for i in number:
                self._channels[i].label = new_value[i]
        else:
            print('number of channels does not match number of labels')

    channelLabels = property(get_channel_labels)

    def get_biosemi_channels(self):
        return self._header['channels']

    def get_channels(self):
        return self._channels

    def set_channels(self, value):
        self._channels = value

    channels = property(get_channels, set_channels)

    def get_system(self):
        return 'Biosemi'

    system = property(get_system)

    def get_amplitude_units(self):
        return self._amplitude_units

    def set_amplitude_units(self, new_value):
        self._amplitude_units = new_value

    amplitude_units = property(get_amplitude_units, set_amplitude_units)

    def get_time_units(self):
        return self._amplitude_units

    def set_time_units(self, new_value):
        self._amplitude_units = new_value

    time_units = property(get_time_units, set_time_units)

    def get_paths(self):
        return self._paths

    paths = property(get_paths)

    def find_triggers(self, **kwargs):
        data = kwargs.get('data', None)
        positive = kwargs.get('positive', True)  # detect only positive
        ideal_epoch_length = kwargs.get('ideal_epoch_length', None)
        fig_file_name = kwargs.get('fig_file_name', None)
        data -= np.mean(data)
        if not positive:
            data *= -1
        data[data <= 0] = 0

        diff_trig = np.diff(np.insert(data, 0, 0))
        #  we will detect positive trantients only
        diff_trig[diff_trig < 0] = 0
        #  we find a detection threshold
        n_triggers = diff_trig.shape[0] / (self._fs * ideal_epoch_length)
        q = 100 - (n_triggers / diff_trig.shape[0]) * 100.0
        h_thr = np.percentile(diff_trig, q)
        _thr = h_thr / 2.0
        diff_trig[diff_trig < _thr] = 0
        # second derivative
        sec_diff = np.diff(np.insert(diff_trig, 0, 0))
        pos = np.where(np.diff(np.sign(sec_diff)))[0]

        # remove positions where second derivative is equal to zero
        pos = pos[sec_diff[pos] != 0]

        c_pos = self.clean_noisy_triggers(trigger_positions=pos,
                                          expected_n_samples=int(self._fs * ideal_epoch_length))
        event_table = {'code': np.ones(c_pos.shape),
                       'idx': c_pos,
                       'dur': np.zeros(c_pos.shape),
                       'dur_samples': np.zeros(c_pos.shape)}
        if fig_file_name is not None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(data)
            ax.plot(c_pos, np.zeros(c_pos.shape), 'o')
            fig.savefig(fig_file_name)
            print('detected triggers saved to {:s}'.format(fig_file_name))

        return event_table

    @staticmethod
    def clean_noisy_triggers(trigger_positions=[], expected_n_samples=None):
        samples_per_trigger = expected_n_samples * 0.9  # 90% tolerance
        cposIn = 0
        cposOut = 1
        triggerPosClean = np.array([])
        badtriggerPos = np.array([])
        while cposOut < len(trigger_positions):
            if trigger_positions[cposOut] - trigger_positions[cposIn] < samples_per_trigger:
                badtriggerPos = np.append(badtriggerPos, trigger_positions[cposOut])
                cposOut = cposOut + 1
            elif cposOut == len(trigger_positions) - 1:
                triggerPosClean = np.append(triggerPosClean, trigger_positions[cposIn], trigger_positions[cposOut])
                cposOut += 1
            else:
                triggerPosClean = np.append(triggerPosClean, trigger_positions[cposIn])
                cposIn = cposOut
                cposOut += 1
        return triggerPosClean.astype(int)



import cv2
import matplotlib.pyplot as plt
import numpy as np


class Recorder(object):
    def __init__(self,
                 video_path: str | None = None,
                 ):
        """
        This class handles the recording of a video which consists of frames generated from the passed matplotlib
        figure by the add_frame method.
        :param video_path: full path to output video
        """
        self.video_path = video_path
        self._frame_counter = 0
        self.__video: cv2.VideoWriter = None

    def add_frame(self, fig: type(plt.figure) | None = None):
        fig.canvas.draw()
        fig.canvas.flush_events()
        img = np.frombuffer(fig.canvas.tostring_argb(), dtype=np.uint8)
        img = img.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        img = np.roll(img, -1, 2)
        img = cv2.cvtColor(img, cv2.COLOR_RGBA2BGR)
        if self._frame_counter == 0:
            # initialize recorder
            fourcc = cv2.VideoWriter_fourcc(*'mp4v')
            height, width, layers = img.shape
            self.__video = cv2.VideoWriter(self.video_path, fourcc, 1, (width, height))
        self.__video.write(img)
        self._frame_counter += 1

    def end_recording(self):
        self.__video.release()
        print('A total of {:} frames were added to {:}'.format(
            self._frame_counter,
            self.video_path))



import copy
from typing import List

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy import units as u
from matplotlib import ticker
from matplotlib.lines import Line2D
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from peegy.definitions.channel_definitions import Domain
from peegy.io.storage.data_storage_reading_tools import (sqlite_tables_to_pandas, sqlite_all_waveforms_to_pandas,
                                                         group_waveforms_df_by)
from peegy.layouts.layouts import Layout
from peegy.plot import eeg_plot_tools as eegpt

__author__ = 'jundurraga'


def has_twin(ax: plt.Axes | None = None):
    """
    Check if input axis has twin axis
    :param ax:
    :return: bool indicating whether twin axis is present or not
    """
    for other_ax in ax.figure.axes:
        if other_ax is ax:
            continue
        if other_ax.bbox.bounds == ax.bbox.bounds:
            return True
    return False


def merge_waveforms_by(df: pd.DataFrame | None = None, group_by: list[str] | None = None):
    """
    This function will pool together the waveforoms from different rows in a dataframe
    :param df: pandas data frame
    :param group_by: grouping keys in dataframe
    :return: pandas dataframe with grouped waveforms
    """
    groups = df.groupby(group_by,
                        observed=False,
                        as_index=False,
                        sort=False)
    output = pd.DataFrame()
    for _id, (_group_keys, _group) in enumerate(groups):
        if not _group.size:
            continue
        _current_group = _group.iloc[[0]].copy()
        _data = np.array([])
        for _, _current_sub_row in _group.iterrows():
            _new_data = _current_sub_row['y']
            if _new_data.ndim == 1:
                _new_data = _new_data[:, None]
            if not _data.size:
                _data = _new_data
            else:
                _data = np.hstack([_data, _new_data])
        print('group {:} output shape {:}'.format(_group_keys, _data.shape))
        _current_group['y'] = _current_group['y'].apply(lambda x: _data)
        output = pd.concat([output, _current_group], ignore_index=True)
    return output


def plot_time_frequency_responses(dataframe: pd.DataFrame | None = None,
                                  rows_by: str | None = None,
                                  cols_by: str | None = None,
                                  color_map: str = 'viridis',
                                  sub_average_time_buffer_size: int | None = None,
                                  time_xlim: type([float, float]) = None,
                                  time_ylim: type([float, float]) = None,
                                  freq_xlim: type([float, float]) = None,
                                  freq_ylim: type([float, float]) = None,
                                  time_vmarkers: type(np.array) | None = None,
                                  freq_vmarkers: type(np.array) | None = None,
                                  freq_vmarkers_style: str | None = None,
                                  show_individual_waveforms: bool = True,
                                  individual_waveforms_alpha: float = 0.1,
                                  show_mean: bool = True,
                                  show_sem: bool = False,
                                  sem_alpha: float = 0.2,
                                  show_sd: bool = True,
                                  sd_alpha: float = 0.2,
                                  show_legend: bool = True,
                                  legend_location='upper center',
                                  title_by: str = 'row',
                                  title_v_offset: float = 0.0,
                                  ylabel: str | None = None,
                                  y_unit_to: u.Unit | None = None,
                                  x_unit_to: u.Unit | None = None,
                                  inset_y_unit_to: u.Unit | None = None,
                                  inset_x_unit_to: u.Unit | None = None,
                                  fontsize: float = 6,
                                  group_by: list[str] | None = None,
                                  colour_by: list[str] | None = None,
                                  show_column_labels: bool = True,
                                  insets_col_source: str | None = None,
                                  insets_col_value: str | None = None,
                                  insets_position: [float, float, float, float] = None,
                                  common_legend: bool = True,
                                  ) -> plt.figure:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data).
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy
    :param rows_by: name of the factor in the dataframe for which the rows in the plot will be split
    :param cols_by: name of the factor in the dataframe for which the columns in the plot will be split
    :param color_map: name of matplotlib colormap (e.g. 'viridis', 'Spectral', 'coolwarm', etcetera
    :param sub_average_time_buffer_size: This is a parameter used to sub_average time_domain data. For example, if each
    of your data have 10000 points, and you want to show the average having a length of 1000 samples, you could specify
    sub_average_time_buffer_size = 1000. This will averaged the 10000 points by splitting the data into blocks of 1000
    samples
    :param time_xlim: x axis limit for the time-domain panels
    :param time_ylim: y axis limit for the time-domain panels
    :param freq_xlim: x axis limit for the frequency-domain panels
    :param freq_ylim: y axis limit for the frequency-domain panels
    :param time_vmarkers: array with x values to add a vertical marker in the time-domain panels
    :param freq_vmarkers: array with x values to add a vertical marker in the frequency-domain panels
    :param show_individual_waveforms: if true, individual waveforms will be shown.
    :param individual_waveforms_alpha: value between 0 and 1 indicating the alpha level of individual waveforms
    :param show_mean: if true, the mean across conditions will be shown
    :param show_sem: if true, the standard error of the mean will be shown
    :param sem_alpha: value between 0 and 1 indicating the alpha level of the sem
    :param show_sd: if true, the standard deviation of the mean will be shown
    :param sd_alpha: value between 0 and 1 indicating the alpha level of the standard deviation
    :param show_legend: if True, the legend for any other category present in the dataframe will be shown
    :param title_by: string specifying from which factor you want to show the titles in each panel. This can be: "row",
    "col", or "both"
    :param title_v_offset: float specifying the vertical offset of the title
    :param freq_vmarkers_style: style of the marker in the frequency-domain. If not passed, vertical lines are used.
    :param ylabel: Label to put in the vertical axis. If empty, the Amplitude and unit of it are used.
    :param y_unit_to: Specify the units in which the vertical axis will be displayed
    :param x_unit_to: Specify the units in which the horizontal axis will be displayed
    :param fontsize: the font size
    :param group_by: list of keys in dataframe which will be grouped and coded by colour
    :param show_column_labels: bool indicatint whether the columns labels should be shown
    :param common_legend: if True, a single common legend will be used. Otherwise, each panel will contain its own
    legend
    :return:
    """
    df = copy.copy(dataframe)
    df.reset_index(inplace=True, drop=True)
    _rows_and_cols = []
    row_conditions = np.array([''])
    col_conditions = np.array([''])
    if rows_by is not None:
        _rows_and_cols.append(rows_by)
        row_conditions = df[rows_by].astype("category").cat.categories
        row_domains = df.iloc[df[rows_by].index]['domain'].unique()
        unique_row_domains = row_domains.size == 1
        n_rows = row_conditions.size
    else:
        df.loc[:, 'dummy_row'] = ''
        _rows_and_cols.append('dummy_row')
        row_domains = df['domain'].unique()
        unique_row_domains = row_domains.size == 1
        n_rows = 1
    idx_rows = np.arange(row_conditions.size)
    if cols_by is not None:
        _rows_and_cols.append(cols_by)
        col_conditions = df[cols_by].astype("category").cat.categories
        n_cols = col_conditions.size
    else:
        df.loc[:, 'dummy_col'] = ''
        _rows_and_cols.append('dummy_col')
        n_cols = 1
    idx_cols = np.arange(col_conditions.size)
    _x_unit_to = x_unit_to
    _y_unit_to = y_unit_to

    # set colurs from colormap
    df['dummy_color_categories'] = ''
    iterate_by = None
    new_var = pd.Series(df.loc[:, ['dummy_color_categories']].values[:, 0])
    if group_by is None and colour_by is None:
        n_colors = 1
        unique_color_factors = np.array([''])
    else:
        group_by = colour_by if group_by is None else group_by
        iterate_by = group_by
        if colour_by is not None:
            iterate_by = colour_by
        sep = '|' if len(iterate_by) > 1 else ''
        for _col_name in iterate_by:
            new_var = new_var + df[_col_name].values.astype(str) + sep
        df['dummy_color_categories'] = new_var
        for _col_name in iterate_by:
            df[_col_name] = pd.Categorical(df[_col_name])
        # sort to keep factor order in colours
        df.sort_values(by=iterate_by, inplace=True)
        vals, idx = np.unique(df['dummy_color_categories'], return_index=True)
        unsorted_unique = vals[np.argsort(idx)]

        df['dummy_color_categories'] = pd.Categorical(df['dummy_color_categories'],
                                                      categories=unsorted_unique)
        unique_color_factors = np.array(df['dummy_color_categories'].cat.categories.tolist())

    n_colors = unique_color_factors.size
    cmap = mpl.colormaps[color_map]
    _colors = cmap(np.linspace(0, 1, n_colors))
    _colors = [mpl.colors.to_hex(cmap(i)) for i in range(n_colors)]

    groups = df.groupby(_rows_and_cols, observed=False, sort=True)
    fig_out = plt.figure(constrained_layout=False)
    # set up subplots
    widths = [1.0] * n_cols
    heights = [1.0] * n_rows
    heights.insert(0, 0.01)  # this is for the top labels
    widths.append(0.01)  # this is for the right labels
    gs = fig_out.add_gridspec(ncols=n_cols + 1, nrows=n_rows + 1,
                              width_ratios=widths,
                              height_ratios=heights,
                              )

    all_ax_row_label = []
    all_ax_col_label = []
    # attach all axis to figure to ensure the are all created before calling them
    [plt.subplot(gs[_idx_row, _idx_col]) for _idx_row in range(gs.get_geometry()[0]) for _idx_col in
     range(gs.get_geometry()[1])]

    for _id, ((_current_row_group, _current_col_group), _group) in enumerate(groups):
        if insets_col_source is None:
            _current_group = _group
        else:
            _current_group = _group[_group[insets_col_source] != insets_col_value]

        if group_by is not None:
            _current_group = merge_waveforms_by(df=_current_group, group_by=group_by)
        if iterate_by is not None:
            _current_group.sort_values(by=iterate_by, inplace=True)
        _inset_df = None
        if insets_col_source is not None and insets_col_value is not None:
            _inset_df = _group[_group[insets_col_source] == insets_col_value]

        _idx_row = idx_rows[_current_row_group == row_conditions].squeeze() + 1
        _idx_col = idx_cols[_current_col_group == col_conditions].squeeze()

        _columns_legends = []
        for _col in _current_group:
            _filter = _current_group[_col].apply(lambda x:
                                                 not (isinstance(x, str) or
                                                      isinstance(x, float) or
                                                      isinstance(x, int)))
            if np.any(_filter.to_numpy()):
                continue
            if colour_by is not None:
                if _col in colour_by:
                    _columns_legends.append(_col)
            else:
                if group_by is not None:
                    if _col in group_by:
                        _columns_legends.append(_col)
                elif len(_current_group[_col].unique()) > 1:
                    _columns_legends.append(_col)

        ax = plt.subplot(gs[_idx_row, _idx_col])
        _last_col_ax = plt.subplot(gs[_idx_row, n_cols])
        for _i, (_, _row) in enumerate(_current_group.iterrows()):
            # get color index
            _color_index,  = np.where(_row['dummy_color_categories'] == unique_color_factors)
            _color_index = int(_color_index)

            _row_summary = add_waveforms_to_df(
                _row=_row,
                _y_unit_to=y_unit_to,
                _x_unit_to=x_unit_to,
                sub_average_time_buffer_size=sub_average_time_buffer_size
            )

            set_axis(ax=ax,
                     _row=_row_summary,
                     _columns_legends=_columns_legends,
                     _current_row_group=_current_row_group,
                     _current_col_group=_current_col_group,
                     _idx_row=_idx_row,
                     time_xlim=time_xlim,
                     time_ylim=time_ylim,
                     freq_xlim=freq_xlim,
                     freq_ylim=freq_ylim,
                     time_vmarkers=time_vmarkers,
                     freq_vmarkers=freq_vmarkers,
                     freq_vmarkers_style=freq_vmarkers_style,
                     show_individual_waveforms=show_individual_waveforms,
                     individual_waveforms_alpha=individual_waveforms_alpha,
                     show_mean=show_mean,
                     show_sem=show_sem,
                     sem_alpha=sem_alpha,
                     show_sd=show_sd,
                     sd_alpha=sd_alpha,
                     show_legend=show_legend,
                     title_by=title_by,
                     title_v_offset=title_v_offset,
                     fontsize=fontsize,
                     color=_colors[_color_index],
                     )
        # all_ax_row_label.append(_last_col_ax)
        # set right labels
        _last_col_ax.set_xticks([])
        _last_col_ax.set_yticks([])
        _last_col_ax.axis('off')
        if not has_twin(_last_col_ax):
            ax_row_label = _last_col_ax.twinx()
            ax_row_label.autoscale(enable=False)
            ax_row_label.set_ylabel(_current_row_group,
                                    size=fontsize,
                                    rotation=-90)
            ax_row_label.set_yticklabels([])
            ax_row_label.tick_params(
                axis='y',  # changes apply to the y-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=True,
                labelsize=fontsize)
            ax_row_label.spines['left'].set_visible(False)
            ax_row_label.spines['right'].set_visible(False)
            ax_row_label.spines['top'].set_visible(False)
            ax_row_label.spines['bottom'].set_visible(False)
            ax_row_label.set_xticks([])
            ax_row_label.set_yticks([])
            all_ax_row_label.append(ax_row_label)
        # set top labels
        _first_row_ax = plt.subplot(gs[0, _idx_col])
        _first_row_ax.set_xticks([])
        _first_row_ax.set_yticks([])
        _first_row_ax.axis('off')
        if not has_twin(_first_row_ax):
            ax_col_label = _first_row_ax.twiny()
            ax_col_label.autoscale(enable=False)
            if show_column_labels:
                ax_col_label.set_xlabel(_current_col_group, size=fontsize)
            ax_col_label.set_xticklabels([])
            ax_col_label.tick_params(
                axis='x',  # changes apply to the x-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=True,
                labelsize=fontsize)
            ax_col_label.spines['left'].set_visible(False)
            ax_col_label.spines['right'].set_visible(False)
            ax_col_label.spines['top'].set_visible(False)
            ax_col_label.spines['bottom'].set_visible(False)
            ax_col_label.set_xticks([])
            ax_col_label.set_yticks([])
            all_ax_col_label.append(ax_col_label)
        if inset_axes and _inset_df is not None:
            if insets_position is None:
                _position = [0.7, 0.8, 0.3, 0.2]
            else:
                _position = tuple(insets_position)

            _insets = get_insets(ax)
            if len(_insets) > 0:
                inset_ax = _insets[0]
            else:
                inset_ax = ax.inset_axes(_position,
                                         transform=ax.transAxes)
            if _inset_df.size:
                _current_inset = merge_waveforms_by(df=_inset_df[_i:_i + 1], group_by=insets_col_source).iloc[0]
                _inset_data = add_waveforms_to_df(
                    _row=_current_inset,
                    _y_unit_to=inset_y_unit_to,
                    _x_unit_to=inset_x_unit_to,
                    sub_average_time_buffer_size=sub_average_time_buffer_size
                )

                set_axis(ax=inset_ax,
                         _row=_inset_data,
                         _columns_legends=_columns_legends,
                         _current_row_group=_current_row_group,
                         _current_col_group=_current_col_group,
                         _idx_row=_idx_row,
                         time_xlim=time_xlim,
                         time_ylim=time_ylim,
                         freq_xlim=freq_xlim,
                         freq_ylim=freq_ylim,
                         time_vmarkers=time_vmarkers,
                         freq_vmarkers=freq_vmarkers,
                         freq_vmarkers_style=freq_vmarkers_style,
                         show_individual_waveforms=show_individual_waveforms,
                         individual_waveforms_alpha=individual_waveforms_alpha,
                         show_sem=show_sem,
                         sem_alpha=sem_alpha,
                         show_sd=show_sd,
                         sd_alpha=sd_alpha,
                         show_legend=show_legend,
                         title_by=title_by,
                         title_v_offset=title_v_offset,
                         fontsize=fontsize,
                         color=_colors[_color_index],
                         )
        if _x_unit_to is None:
            _x_unit_to = _row_summary['x_unit']
        if _y_unit_to is None:
            _y_unit_to = _row_summary['y_unit']
    all_axes = fig_out.get_axes()
    axis_with_data = [ax for ax in all_axes if
                      ax not in all_ax_row_label and
                      ax not in all_ax_col_label and
                      ax.has_data()]
    all_axis_with_data = np.all(axis_with_data)
    # check if there is a common legend. If so, then we keep only one across all axes
    _unique_labels = []
    _unique_handles = []
    for _ax in axis_with_data:
        _handles, _labels = _ax.get_legend_handles_labels()
        if len(_handles) > 0 and len(_labels) > 0:
            _unique_handles = _unique_handles + [_handles]
            _unique_labels = _unique_labels + [_labels]
    _unique_handles = [b for b in _unique_handles if b]
    _unique_labels = [b for b in _unique_labels if b]

    # common_labels = True
    # for _idx_ul in range(len(_unique_labels) - 1):
    #     diff = set(_unique_labels[_idx_ul]) - set(_unique_labels[_idx_ul + 1])
    #     if len(diff) > 0:
    #         common_labels = False
    #         break
    if common_legend and len(_unique_handles) > 0:
        for _ax in axis_with_data:
            if _ax.legend_ is not None:
                _ax.get_legend().remove()
        legend_elements = [
            Line2D([0],
                   [0],
                   color=_color,
                   lw=1,
                   label=_label) for _color, _label in zip(_colors,
                                                           unique_color_factors)]
        fig_out.legend(handles=legend_elements,
                       loc=legend_location,
                       fontsize=fontsize,
                       frameon=False,
                       ncol=len(legend_elements),
                       handlelength=1
                       )

    if ylabel is None:
        if _y_unit_to is not None:
            fig_out.supylabel('Amplitude [{:}]'.format(_y_unit_to), size=fontsize)
    else:
        fig_out.supylabel(ylabel, size=fontsize)
    if unique_row_domains and all_axis_with_data and np.all(row_domains == Domain.time) and n_cols > 1:
        fig_out.supxlabel('Time [{:}]'.format(_x_unit_to), size=fontsize)
    if unique_row_domains and all_axis_with_data and np.all(row_domains == Domain.frequency) and n_cols > 1:
        fig_out.supxlabel('Frequency [{:}]'.format(_x_unit_to), size=fontsize)

    unique_x_lim_per_row = []
    for _i_col in range(gs.get_geometry()[1] - 1):
        all_axis = [plt.subplot(gs[_i_row, _i_col]) for _i_row in range(1, gs.get_geometry()[0])]
        lims = []
        for _ax in all_axis:
            lims.append(np.array(_ax.get_xlim()))
        unique_x_lim_per_row.append(np.unique(np.array(lims), axis=0).shape[0] == 1)
    unique_x_lim = np.all(unique_x_lim_per_row)

    unique_y_lim_per_col = []
    for _i_row in range(1, gs.get_geometry()[0]):
        all_axis = [plt.subplot(gs[_i_row, _i_col]) for _i_col in range(gs.get_geometry()[1] - 1)]
        lims = []
        for _ax in all_axis:
            if _ax.has_data():
                lims.append(np.array(_ax.get_ylim()))
        unique_y_lim_per_col.append(np.unique(np.array(lims), axis=0).shape[0] == 1)
    unique_y_lim = np.all(unique_y_lim_per_col)
    for ax in all_axes:
        if not ax.has_data() and ax not in all_ax_row_label and ax not in all_ax_col_label:
            ax.set_visible(False)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.set_xlabel('')
            continue
        ax.spines['top'].set_visible(False)
        if ax not in all_ax_col_label:
            if unique_row_domains and all_axis_with_data:
                subplotspec = ax.get_subplotspec()
                if subplotspec is not None and not subplotspec.is_last_row() and ax.has_data():
                    ax.set_xticklabels([])
                    ax.set_xlabel('')
                if n_cols > 1:
                    ax.set_xlabel('')
            elif unique_x_lim:
                subplotspec = ax.get_subplotspec()
                if subplotspec is not None and not subplotspec.is_last_row() and ax.has_data():
                    ax.set_xticklabels([])
                    ax.set_xlabel('')

            if unique_y_lim:
                subplotspec = ax.get_subplotspec()
                if subplotspec is not None and not subplotspec.is_first_col() and ax.has_data():
                    ax.set_yticklabels([])
                    ax.set_ylabel('')

        # ax.spines['left'].set_visible(True)
        ax.spines['right'].set_visible(False)
        ax.tick_params(labelsize=fontsize)

    inch = 2.54
    fig_out.set_size_inches(12.0 / inch, 2.25 * len(row_conditions) / inch)
    fig_out.subplots_adjust(top=0.98, bottom=0.08, hspace=0.0, left=0.15, right=0.95)
    fig_out.tight_layout()
    return fig_out


def plot_topographic_maps(dataframe: pd.DataFrame | None = None,
                          rows_by: str | None = None,
                          cols_by: str | None = None,
                          subject_id_column: str = 'subject_id',
                          normalize_by: str = 'subject_id',
                          channels_column: str = 'channel',
                          title: str = '',
                          topographic_value: str | None = None,
                          layout: str | None = None,
                          title_by: str | None = None,
                          title_v_offset: float = 0.0,
                          fontsize: float = 6,
                          grid_size: np.complex128 = 600j,
                          color_map_label: str | None = None,
                          normalize: bool = False,
                          show_sensors: bool = True,
                          show_sensor_label: bool = False,
                          n_contour_levels: int = 35,
                          max_topographic_value: float | None = None,
                          min_topographic_value: float | None = None,
                          contour_line_width: float = 0.1,
                          apply_function: object = np.mean,
                          fun_args: dict = {'axis': 2}
                          ) -> plt.figure:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data)
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy
    :param rows_by: name of the factor in the dataframe for which the rows in the plot will be split
    :param cols_by: name of the factor in the dataframe for which the columns in the plot will be split
    :param subject_id_column: string indicating the column name with subject ids
    :param normalize_by: string indicating the column that will be used to normalized the data
    :param channels_column: name of column containing channel labels
    :param title: title of the figure
    :param topographic_value: name of column containing the value to be shown by the topographic map
    :param layout: path or name of the layout to be used
    :param title_by: string specifying from which factor you want to show the titles in each panel. This can be: "row",
    "col", or "both"
    :param title_v_offset: float specifying the vertical offset of the title
    :param fontsize: the fontsize
    :param grid_size: complex number indicating the size of the grid,
    :param color_map_label: string with the label that would be use of the colourmap. If empty, the default value will
    be the topographic_value
    :param normalize: if True, topographic maps will be normalized within normalize_by
    :param show_sensors: if True, the position of the sensors will be shown
    :param show_sensor_label: if True, the label of the sensor will be shown
    :param n_contour_levels: number of contour lines to plot
    :param min_topographic_value: if given, the colour scale will be restricted to have this value as maximum, otherwise
    it will be determined from the data.
    :param max_topographic_value: if given, the colour scale will be restricted to have this value as minimum, otherwise
    it will be determined from the data.
    :param contour_line_width: float indicating the with (in points) of the contour lines
    :param apply_function: function to be applied to each subgroup
    :param fun_args: extra arguments to be used by apply_function. This is a dictionary with the desired parameters
    :return:
    """
    df = copy.copy(dataframe)
    _rows_and_cols = []
    row_conditions = np.array([''])
    col_conditions = np.array([''])
    if rows_by is not None:
        _rows_and_cols.append(rows_by)
        row_conditions = df[rows_by].astype("category").cat.categories
        n_rows = row_conditions.size
    else:
        df.loc[:, 'dummy_row'] = ''
        _rows_and_cols.append('dummy_row')
        n_rows = 1
    idx_rows = np.arange(row_conditions.size)

    if cols_by is not None:
        _rows_and_cols.append(cols_by)
        col_conditions = df[cols_by].astype("category").cat.categories
        n_cols = col_conditions.size
    else:
        df.loc[:, 'dummy_col'] = ''
        _rows_and_cols.append('dummy_col')
        n_cols = 1
    idx_cols = np.arange(col_conditions.size)
    # check if input dataframe already contains the potentials. If not, then we extract them
    if 'potentials' in df.keys():
        if normalize:
            if normalize_by is not None:
                df = df.assign(
                    potentials=df.groupby(
                        normalize_by, observed=False)['potentials'].transform(
                        lambda x: x / np.ma.max(np.abs(np.ma.dstack([_val for _val in x])))))
            else:
                df = df.assign(
                    potentials=df.groupby(
                        [True] * len(df))['potentials'].transform(
                        lambda x: x / np.ma.max(np.abs(np.ma.dstack([_val for _val in x])))))
        sub_groups = df
    else:
        df['topo_value'] = df[topographic_value]
        if normalize:
            df = df.assign(
                topo_value=df.groupby(
                    [normalize_by], observed=False)[topographic_value].transform(lambda x: x / np.max(np.abs(x))))
        sub_groups = get_topographic_maps_as_df(dataframe=df,
                                                subject_id_column=subject_id_column,
                                                group_by=_rows_and_cols,
                                                channels_column='channel',
                                                topographic_value='topo_value',
                                                topographic_label=topographic_value,
                                                layout=layout,
                                                apply_function=apply_function,
                                                fun_args=fun_args,
                                                grid_size=grid_size,
                                                )
    if color_map_label is None:
        color_map_label = topographic_value
    groups = sub_groups.groupby(_rows_and_cols, observed=False)
    fig_out = plt.figure(constrained_layout=True)
    widths = [1.0] * n_cols
    heights = [1.0] * n_rows
    heights.append(0.1)  # this is for the color bar on the bottom
    heights.insert(0, 0.1)  # this is for the top labels
    widths.append(0.01)  # this is for the right labels
    gs = fig_out.add_gridspec(ncols=n_cols + 1, nrows=n_rows + 2,
                              width_ratios=widths,
                              height_ratios=heights,
                              )
    if max_topographic_value is None:
        max_sub_groups = sub_groups[sub_groups.potentials.apply(lambda x: np.size(x) > 1)]
        max_sub = max_sub_groups.potentials.apply(
            lambda x: np.ma.dstack([_val for _val in x]).max()).max()
    else:
        max_sub = max_topographic_value

    if min_topographic_value is None:
        min_sub_groups = sub_groups[sub_groups.potentials.apply(lambda x: np.size(x) > 1)]
        min_sub = min_sub_groups.potentials.apply(
            lambda x: np.ma.dstack([_val for _val in x]).min()).min()
    else:
        min_sub = min_topographic_value
    max_distance = sub_groups.max_distance.max()
    all_ax_row_label = []
    all_ax_col_label = []
    for _id, ((_current_row_group, _current_col_group), _group) in enumerate(groups):
        if _group.shape[0] == 0 or np.all(np.isnan(_group.potentials.values[0])):
            continue
        _idx_row = idx_rows[_current_row_group == row_conditions].squeeze() + 1
        _idx_col = idx_cols[_current_col_group == col_conditions].squeeze()
        ax = plt.subplot(gs[_idx_row, _idx_col])
        panel_title = ''
        if title_by == 'row':
            panel_title = '{:}'.format(_current_row_group)
        if title_by == 'col':
            panel_title = '{:}'.format(_current_col_group)

        if title_by == 'both':
            panel_title = '{:} / {:}'.format(_current_row_group, _current_col_group)

        ax.set_title(panel_title,
                     y=1 + title_v_offset,
                     size=fontsize,
                     fontweight='bold')
        if _group.potentials.values.size > 1:
            std_average = _group.potentials.apply(
                lambda x: apply_function(np.ma.dstack([_val for _val in x]), *fun_args)).values[0]
        else:
            std_average = _group.potentials.values[0]
        ax_im = ax.imshow(std_average.T, origin='lower',
                          extent=(-max_distance, max_distance, -max_distance, max_distance),
                          vmin=min_sub,
                          vmax=max_sub,
                          aspect=1.0)
        ax_im.set_cmap('nipy_spectral')

        levels = np.linspace(min_sub, max_sub, n_contour_levels)
        ax.contour(std_average.T,
                   levels,
                   origin='lower',
                   extent=(-max_distance, max_distance, -max_distance, max_distance),
                   linewidths=contour_line_width,
                   colors='k',
                   linestyles='solid')
        ax.autoscale(enable=False)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis('off')
        # set right labels
        _last_col_ax = plt.subplot(gs[_idx_row, n_cols])
        _last_col_ax.set_xticks([])
        _last_col_ax.set_yticks([])
        _last_col_ax.axis('off')
        if not has_twin(_last_col_ax):
            ax_row_label = _last_col_ax.twinx()
            ax_row_label.autoscale(enable=False)
            ax_row_label.set_ylabel(_current_row_group,
                                    size=fontsize,
                                    rotation=-90)
            ax_row_label.set_yticklabels([])
            ax_row_label.tick_params(
                axis='y',  # changes apply to the y-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=True,
                labelsize=fontsize)
            ax_row_label.spines['left'].set_visible(False)
            ax_row_label.spines['right'].set_visible(False)
            ax_row_label.spines['top'].set_visible(False)
            ax_row_label.spines['bottom'].set_visible(False)
            ax_row_label.set_xticks([])
            ax_row_label.set_yticks([])
            all_ax_row_label.append(ax_row_label)
        # set top labels
        _first_row_ax = plt.subplot(gs[0, _idx_col])
        _first_row_ax.set_xticks([])
        _first_row_ax.set_yticks([])
        _first_row_ax.axis('off')
        if not has_twin(_first_row_ax):
            ax_col_label = _first_row_ax.twiny()
            ax_col_label.autoscale(enable=False)
            ax_col_label.set_xlabel(_current_col_group, size=fontsize)
            ax_col_label.set_xticklabels([])
            ax_col_label.tick_params(
                axis='x',  # changes apply to the x-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=True,
                labelsize=fontsize)
            ax_col_label.spines['left'].set_visible(False)
            ax_col_label.spines['right'].set_visible(False)
            ax_col_label.spines['top'].set_visible(False)
            ax_col_label.spines['bottom'].set_visible(False)
            ax_col_label.set_xticks([])
            ax_col_label.set_yticks([])
            all_ax_col_label.append(ax_col_label)
        # plot color bar
        if _id == 0:
            c_bar_ax = plt.subplot(gs[-1, :])
            c_bar = fig_out.colorbar(ax_im, cax=c_bar_ax, orientation='horizontal', format='%.1f')
            c_bar.set_label(color_map_label, fontsize=fontsize)
            tick_locator = ticker.MaxNLocator(nbins=3)
            c_bar.locator = tick_locator
            c_bar.update_ticks()
            c_bar.ax.tick_params(labelsize=fontsize)
        channels = _group.channels.values[0]
        channel_labels = [ch['label'] for ch in channels]
        ax.plot(0, max_distance * 1.0, '|', markersize=8, color='k')
        if show_sensors:
            _lay = Layout()
            _layout = _lay.get_layout(file_name=layout)
            _layout = np.array([_l for _l in _layout if _l.label not in ['COMNT', 'SCALE']])
            for i, lay in enumerate(_layout):
                if lay.label in channel_labels:
                    ax.plot(lay.x, lay.y, 'o', color='b', markersize=1)
                else:
                    ax.plot(lay.x, lay.y, 'o', color='grey', markersize=1)
                if show_sensor_label:
                    ax.text(lay.x, lay.y, s=lay.label, fontsize=6)

    all_axes = fig_out.get_axes()
    for ax in all_axes:
        if ax == c_bar.ax:
            continue
        ax.spines['top'].set_visible(False)
        # ax.set_xticklabels([])
        # ax.set_xlabel('')
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        # if title_by is not None and not ax.get_subplotspec().is_first_row():
        #     ax.set_title('')
    inch = 2.54
    fig_out.suptitle(title)
    fig_out.set_size_inches(3.2 * len(col_conditions) / inch, h=3.2 * len(row_conditions) / inch)
    return fig_out


def get_topographic_maps(df,
                         subject_id_column: str = 'subject_id',
                         channels_column: str = 'channel',
                         topographic_value: str | None = None,
                         layout: str | None = None,
                         grid_size: np.complex128 = 600j,
                         apply_function: object | None = None,
                         fun_args: dict = {}):
    _lay = Layout()
    _layout = _lay.get_layout(file_name=layout)
    _layout = np.array([_l for _l in _layout if _l.label not in ['COMNT', 'SCALE']])
    _single_responses = np.array([])

    df[subject_id_column] = df[subject_id_column].astype('category').cat.remove_unused_categories()
    _subject_groups = df.groupby(subject_id_column, observed=False, group_keys=False)
    for _, (_id_sub, _sub_group) in enumerate(_subject_groups):
        if _sub_group.size == 0:
            print('no data found for {:}'.format(_id_sub))
            continue
        if _sub_group.shape[0] < 4:
            print('no enough data points to generate topographic map ({:} points)'.format(_sub_group.shape[0]))
            continue
        x = []
        y = []
        z = []
        channels = []
        max_potential = -np.inf
        print('Using {:} data points to generate topographic map from layout with {:} sensors'.format(
            _sub_group.shape[0],
            _layout.size
        ))
        for _i, (_, _row) in enumerate(_sub_group.iterrows()):
            ch = _row[channels_column]
            amp = _row[topographic_value]
            for _l in _layout:
                if _l.label.lower() == ch.lower():
                    x.append(_l.x)
                    y.append(_l.y)
                    z.append(amp)
                    channels.append({'label': _l.label})

        _potentials, max_distance = eegpt.interpolate_potential_fields(x=np.array(x).reshape(-1, 1),
                                                                       y=np.array(y).reshape(-1, 1),
                                                                       z=np.array(z).reshape(-1, 1),
                                                                       grid=grid_size)
        max_potential
        if not _single_responses.size:
            _single_responses = _potentials[:, :, None]
        else:
            _single_responses = np.ma.dstack((_single_responses, _potentials))
    out = None
    if _single_responses.size:
        if apply_function is not None:
            _single_responses = apply_function(_single_responses, **fun_args)
        out = pd.Series(data={'potentials': _single_responses,
                              'max_distance': max_distance,
                              'channels': channels
                              })
    return out


def get_topographic_maps_as_df(dataframe: pd.DataFrame | None = None,
                               group_by: List[str] | None = None,
                               subject_id_column: str = 'subject_id',
                               channels_column: str = 'channel',
                               topographic_value: str | None = None,
                               topographic_label: str | None = None,
                               layout: str | None = None,
                               grid_size: np.complex128 = 600j,
                               apply_function: object = np.mean,
                               fun_args: dict = {'axis': 2}
                               ) -> pd.DataFrame:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data).
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy.
    :param group_by: name of factors to group and obtain potentials.
    :param subject_id_column: string indicating the column name with subject ids
    :param channels_column: name of column containing channel labels
    :param topographic_value: name of column containing the value to be shown by the topographic map
    :param topographic_label: name of of the value to being mapped
    :param layout: path or name of the layout to be used
    :param grid_size: complex number indicating the size of the grid,
    :return:
    """
    df = copy.copy(dataframe)
    df['topo_value'] = df[topographic_value]
    if group_by is not None:
        for _g in group_by:
            df[_g] = df[_g].astype('category').cat.remove_unused_categories()
    else:
        group_by = [True] * len(df)
    sub_groups = df.groupby(group_by, group_keys=False, observed=True).apply(
        lambda x: get_topographic_maps(x,
                                       subject_id_column=subject_id_column,
                                       layout=layout,
                                       channels_column=channels_column,
                                       topographic_value='topo_value',
                                       grid_size=grid_size,
                                       apply_function=apply_function,
                                       fun_args=fun_args,
                                       ),
        include_groups=False).reset_index()
    sub_groups['topographic_label'] = topographic_label

    return sub_groups


def set_axis(ax: type(plt.axis) | None = None,
             _row: pd.DataFrame | None = None,
             _columns_legends: str | None = None,
             _y_unit_to: u.Unit | None = None,
             _x_unit_to: u.Unit | None = None,
             sub_average_time_buffer_size: int | None = None,
             # _last_col_ax: plt.Axes | None = None,
             _current_row_group: str | None = None,
             _current_col_group: str | None = None,
             _idx_row: int | None = None,
             _domain: str | None = None,
             time_xlim: [float, float] = None,
             time_ylim: [float, float] = None,
             freq_xlim: [float, float] = None,
             freq_ylim: [float, float] = None,
             time_vmarkers: type(np.array) | None = None,
             freq_vmarkers: type(np.array) | None = None,
             freq_vmarkers_style: str | None = None,
             show_individual_waveforms: bool = True,
             individual_waveforms_alpha: float = 0.1,
             show_mean: bool = True,
             show_sem: bool = False,
             sem_alpha: float = 0.2,
             show_sd: bool = True,
             sd_alpha: float = 0.2,
             show_legend: bool = True,
             title_by: str = 'row',
             title_v_offset: float = 0.0,
             fontsize: float = 6,
             color: [float, float, float] = None,
             _group_label: str | None = None,
             ):
    _group_label = []
    for _ic, _col in enumerate(_columns_legends):
        _group_label.append(str(_row[_col]))
    _group_label = '/'.join(_group_label)
    _domain = _row['domain']

    title = ''
    if title_by == 'row':
        title = '{:}'.format(_current_row_group)

    if title_by == 'col':
        title = '{:}'.format(_current_col_group)

    if title_by == 'both':
        title = '{:} / {:}'.format(_current_row_group, _current_col_group)

    if title_by == 'col':
        if _idx_row == 0:
            ax.set_title(title,
                         y=1 + title_v_offset,
                         size=fontsize,
                         fontweight='bold')
    else:
        ax.set_title(title,
                     y=1 + title_v_offset,
                     size=fontsize,
                     fontweight='bold')

    if show_individual_waveforms:
        ax.plot(_row['x'].value, _row['y_single_responses'].value,
                linewidth=0.5,
                alpha=individual_waveforms_alpha,
                color=color
                )
        if 'min_time' in _row.keys() and 'min_time_unit' in _row.keys():
            _peak_time = _row['min_time'] * u.Quantity(1, _row['min_time_unit']).to(_row['x'].unit)
            _idx_peak = np.argmin(np.abs(_row['x'] - _peak_time))
            ax.plot(_peak_time, _row['y_single_responses'][[_idx_peak], :],
                    markersize=3,
                    alpha=individual_waveforms_alpha,
                    color=color,
                    marker='v'
                    )
    if show_sem:
        ax.fill_between(_row['x'].value,
                        (_row['y_mean'] - _row['y_sem']).value * (_domain == Domain.time),
                        (_row['y_mean'] + _row['y_sem']).value,
                        alpha=sem_alpha,
                        edgecolor="none",
                        facecolor=color)
    if show_sd:
        ax.fill_between(_row['x'].value,
                        (_row['y_mean'] - _row['y_sd']).value * (_domain == Domain.time),
                        (_row['y_mean'] + _row['y_sd']).value,
                        alpha=sd_alpha,
                        edgecolor="none",
                        facecolor=color)
    if show_mean:
        ax.plot(_row['x'].value, _row['y_mean'].value,
                color=color,
                linewidth=0.8,
                label=_group_label)
        if 'min_time' in _row.keys() and 'min_time_unit' in _row.keys():
            _peak_time = _row['min_time'] * u.Quantity(1, _row['min_time_unit']).to(_row['x'].unit)
            _idx_peak = np.argmin(np.abs(_row['x'] - _peak_time))
            ax.plot(_peak_time.value, _row['y_mean'].value[_idx_peak],
                    markersize=3,
                    color=color,
                    marker='v'
                    )
        if 'max_time' in _row.keys() and 'max_time_unit' in _row.keys():
            _peak_time = _row['max_time'] * u.Quantity(1, _row['max_time_unit']).to(_row['x'].unit)
            _idx_peak = np.argmin(np.abs(_row['x'] - _peak_time))
            ax.plot(_peak_time.value, _row['y_mean'].value[_idx_peak],
                    markersize=3,
                    color=color,
                    marker='v'
                    )

    if _domain == Domain.time:
        if time_xlim is not None:
            ax.set_xlim(time_xlim)
        if time_ylim is not None:
            ax.set_ylim(time_ylim)
        if time_vmarkers is not None:
            [ax.axvline(_t, color='k', linestyle=':', linewidth=0.5) for _t in time_vmarkers]
        ax.set_xlabel('Time [{:}]'.format(_row['x'].unit), size=fontsize)

    if _domain == Domain.frequency:
        if freq_xlim is not None:
            ax.set_xlim(freq_xlim)
            if freq_ylim is None:
                _x_idx = np.where(np.logical_and(_row['x'].value >= np.min(freq_xlim),
                                                 _row['x'].value <= np.max(freq_xlim)))[0]
                _freq_ylim = [0, np.max(_row['y_mean'].value[_x_idx]) * 1.2]
                ax.set_ylim(_freq_ylim)

        if freq_ylim is not None:
            ax.set_ylim(freq_ylim)
        if freq_vmarkers is not None:
            if freq_vmarkers_style is None:
                [ax.axvline(_f, color='k', linestyle=':', linewidth=0.5) for _f in freq_vmarkers]
            else:
                y_min, y_max = ax.get_ylim()
                [ax.plot(_f, y_max * 0.93,
                         color='g',
                         marker=freq_vmarkers_style,
                         markersize=3) for _f in freq_vmarkers]
        ax.set_xlabel('Frequency [{:}]'.format(_row['x'].unit), size=fontsize)
    handles, labels = ax.get_legend_handles_labels()
    if len(labels) > 1 and show_legend:
        ax.legend(handles,
                  labels,
                  loc='upper center',
                  fontsize=fontsize,
                  frameon=False,
                  # ncol=len(labels)
                  )
    # return all_ax_row_label
    return


def add_waveforms_to_df(_row: pd.core.series.Series | None = None,
                        _y_unit_to: u.Unit | None = None,
                        _x_unit_to: u.Unit | None = None,
                        sub_average_time_buffer_size: int | None = None,
                        ):
    _series = _row.copy()
    y_unit = u.Quantity(1, _series['y_unit'])
    x_unit = u.Quantity(1, _series['x_unit'])
    if _y_unit_to is None:
        _y_unit_to = y_unit.unit
    if _x_unit_to is None:
        _x_unit_to = x_unit.unit
    y = (_series['y'] * y_unit).to(_y_unit_to)
    x = (_series['x'] * x_unit).to(_x_unit_to)
    fs = _series['x_fs']

    _domain = _series['domain']
    if y.ndim == 1:
        y = y.reshape([-1, 1])
    y_single_responses = y

    if _domain == Domain.time and sub_average_time_buffer_size is not None:
        fs = 1 / np.mean(np.diff(x))
        used_samples = int(np.floor(y.shape[0] // sub_average_time_buffer_size) * sub_average_time_buffer_size)
        y_f = y[0: used_samples, :]
        y_f = np.transpose(np.reshape(y_f, (sub_average_time_buffer_size, -1, y_f.shape[1]), order='F'),
                           [0, 2, 1])
        # demean
        y_f = np.mean(y_f, axis=2)
        x = np.arange(0, sub_average_time_buffer_size) / fs
        y_single_responses = y_f - np.mean(y_f, axis=0)
    y_mean = None
    y_sd = None
    y_sem = None
    if _domain == Domain.time:
        y_mean = np.mean(y_single_responses, axis=1)
        y_sd = np.std(y_single_responses, axis=1)
        y_sem = y_sd / np.sqrt(y_single_responses.shape[1])
    if _domain == Domain.frequency:
        y_mean = np.abs(np.mean(y_single_responses, axis=1))
        y_sd = np.std(np.abs(y_single_responses), axis=1)
        y_sem = y_sd / np.sqrt(y_single_responses.shape[1])
        y_single_responses = np.abs(y_single_responses)
    _series['y_single_responses'] = y_single_responses
    _series['y_mean'] = y_mean
    _series['y_sd'] = y_sd
    _series['y_sem'] = y_sem
    _series['x'] = x
    return _series


def get_insets(ax: plt.Axes):
    return [c for c in ax.get_children()
            if isinstance(c, plt.Axes)]


def get_all_topographic_maps(dataframe: pd.DataFrame | None = None,
                             channels_column: str = 'channel',
                             topographic_value: str | None = None,
                             layout: str | None = None,
                             group_by: List[str] = [],
                             grid_size: np.complex128 = 600j,
                             ) -> pd.DataFrame:
    df = dataframe.copy()
    df['topo_value'] = df[topographic_value]
    df['channel'] = df[channels_column]
    _group_by = ['subject_id',
                 'id_measurement',
                 'id_stimuli',
                 'id_recording'] + group_by
    _group_by = [_g for _g in _group_by if _g != 'channel']
    # extract unique group values
    _group_by = list(sorted(set(_group_by).intersection(set(dataframe.columns)), key=lambda x: _group_by.index(x)))
    sub_groups = get_topographic_maps_as_df(
        dataframe=df,
        subject_id_column='id_subject',
        group_by=_group_by,
        channels_column='channel',
        topographic_value='topo_value',
        topographic_label=topographic_value,
        layout=layout,
        grid_size=grid_size,
    )
    # clean in case data could not be interpolated because of insufficient number of points
    sub_groups = pd.DataFrame.dropna(sub_groups, axis=0)
    sum_df = df.loc[df[_group_by].drop_duplicates().index]
    sum_df = sum_df.drop(columns=['topo_value', 'channel'])
    # output = pd.concat([sum_df.set_index(_group_by),
    #                     sub_groups.set_index(_group_by)],
    #                    axis=1,
    #                    join='inner').reset_index()
    _group_by = list(set(_group_by).intersection(sum_df.keys()))
    _group_by = list(set(_group_by).intersection(sub_groups.keys()))
    output = pd.merge(sum_df, sub_groups,
                      left_on=_group_by,
                      right_on=_group_by,
                      how='outer')

    return output


def get_topographic_maps_group_by(dataframe: pd.DataFrame | None = None,
                                  group_by: List[str] | None = None,
                                  subject_id_column: str = 'subject_id',
                                  normalize_by: str = 'subject_id',
                                  channels_column: str = 'channel',
                                  topographic_value: str | None = None,
                                  layout: str | None = None,
                                  normalize: bool = False,
                                  apply_function: object = np.mean,
                                  fun_args: dict = {'axis': 2}
                                  ) -> pd.DataFrame:
    df = copy.copy(dataframe)
    # check if input dataframe already contains the potentials. If not, then we extract them
    df['topo_value'] = df[topographic_value]
    if normalize:
        df = df.assign(
            topo_value=df.groupby(
                [normalize_by], observed=False)[topographic_value].transform(lambda x: x / np.max(np.abs(x))))

    sub_groups = get_topographic_maps_as_df(dataframe=df,
                                            subject_id_column=subject_id_column,
                                            group_by=group_by,
                                            channels_column='channel',
                                            topographic_value='topo_value',
                                            topographic_label=topographic_value,
                                            layout=layout,
                                            apply_function=apply_function,
                                            fun_args=fun_args
                                            )
    return sub_groups


def group_topographic_df_by(dataframe: type(pd.DataFrame) | None = None,
                            group_by: List[str] | None = None):
    group_by = list(set(group_by).intersection(dataframe.keys()))
    _subject_groups = dataframe.groupby(group_by, observed=False, group_keys=False)
    max_distance = -np.inf
    _unique_channels = []
    out_df = pd.DataFrame(columns=group_by + ['potentials', 'max_distance', 'channels', 'topographic_label'])
    for _, (_id_subgroup, _sub_group) in enumerate(_subject_groups):
        if _sub_group.size == 0:
            print('no data found for {:}'.format(_id_subgroup))
            continue
        _id_subgroup_str = [str(_val) for _val in _id_subgroup]
        print('topographic map group ' + '|'.join(_id_subgroup_str) + ': n = {:}'.format(_sub_group.shape[0]))
        _potentials = _sub_group['potentials']
        max_distance = np.maximum(max_distance, np.max(_sub_group['max_distance']))
        _single_responses = np.array([])
        for _p in _potentials:
            if not _single_responses.size:
                _single_responses = _p[:, :, None]
            else:
                _single_responses = np.ma.dstack((_single_responses, _p))
            if not len(_unique_channels):
                _unique_channels = set([_ch['label'] for _, _row in _sub_group['channels'].items() for _ch in _row])
            else:
                _unique_channels = set(_unique_channels).union(
                    set([_ch['label'] for _, _row in _sub_group['channels'].items() for _ch in _row]))

        _new_df = pd.DataFrame([list(_id_subgroup) + [_single_responses] + [None, None, None]],
                               columns=out_df.columns)
        if out_df.shape[0] == 0:
            out_df = _new_df
        else:
            out_df = pd.concat([_new_df, out_df], ignore_index=True)

    out_df['channels'] = out_df['channels'].apply(lambda x: [{'label': _value} for _value in _unique_channels])
    out_df['max_distance'] = max_distance
    out_df['topographic_label'] = '/'.join(dataframe['topographic_label'].astype(str).unique())
    return out_df


def plot_time_frequency_responses_and_topographic_maps(dataframe: pd.DataFrame | None = None,
                                                       rows_by: str | None = None,
                                                       cols_by: str | None = None,
                                                       color_map: str = 'viridis',
                                                       sub_average_time_buffer_size: int | None = None,
                                                       time_xlim: type([float, float]) = None,
                                                       time_ylim: type([float, float]) = None,
                                                       freq_xlim: type([float, float]) = None,
                                                       freq_ylim: type([float, float]) = None,
                                                       time_vmarkers: type(np.array) | None = None,
                                                       freq_vmarkers: type(np.array) | None = None,
                                                       freq_vmarkers_style: str | None = None,
                                                       show_individual_waveforms: bool = True,
                                                       individual_waveforms_alpha: float = 0.1,
                                                       show_mean: bool = True,
                                                       show_sem: bool = False,
                                                       sem_alpha: float = 0.2,
                                                       show_sd: bool = True,
                                                       sd_alpha: float = 0.2,
                                                       show_legend: bool = True,
                                                       legend_location='upper center',
                                                       title_by: str = 'row',
                                                       title_v_offset: float = 0.0,
                                                       ylabel: str | None = None,
                                                       y_unit_to: u.Unit | None = None,
                                                       x_unit_to: u.Unit | None = None,
                                                       inset_y_unit_to: u.Unit | None = None,
                                                       inset_x_unit_to: u.Unit | None = None,
                                                       fontsize: float = 6,
                                                       group_by: list[str] | None = None,
                                                       colour_by: list[str] | None = None,
                                                       show_column_labels: bool = True,
                                                       insets_position: type([float, float, float, float]) = None,
                                                       color_map_label: str | None = None,
                                                       colormap_size: float = 0.1,
                                                       normalize: bool = False,
                                                       show_sensors: bool = True,
                                                       show_sensor_label: bool = False,
                                                       n_contour_levels: int = 35,
                                                       max_topographic_value: float | None = None,
                                                       min_topographic_value: float | None = None,
                                                       commom_topographic_value: bool = True,
                                                       contour_line_width: float = 0.1,
                                                       apply_function: object = np.nanmean,
                                                       fun_args: dict = {'axis': 2},
                                                       topographic_maps_direction: str = 'v',
                                                       inset_scale: float = 1,
                                                       common_legend: bool = True,
                                                       ) -> plt.figure:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data).
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy
    :param rows_by: name of the factor in the dataframe for which the rows in the plot will be split
    :param cols_by: name of the factor in the dataframe for which the columns in the plot will be split
    :param color_map: name of matplotlib colormap (e.g. 'viridis', 'Spectral', 'coolwarm', etcetera
    :param sub_average_time_buffer_size: This is a parameter used to sub_average time_domain data. For example, if each
    of your data have 10000 points, and you want to show the average having a length of 1000 samples, you could specify
    sub_average_time_buffer_size = 1000. This will averaged the 10000 points by splitting the data into blocks of 1000
    samples
    :param time_xlim: x axis limit for the time-domain panels
    :param time_ylim: y axis limit for the time-domain panels
    :param freq_xlim: x axis limit for the frequency-domain panels
    :param freq_ylim: y axis limit for the frequency-domain panels
    :param time_vmarkers: array with x values to add a vertical marker in the time-domain panels
    :param freq_vmarkers: array with x values to add a vertical marker in the frequency-domain panels
    :param show_individual_waveforms: if true, individual waveforms will be shown.
    :param individual_waveforms_alpha: value between 0 and 1 indicating the alpha level of individual waveforms
    :param show_mean: if true, the mean across conditions will be shown
    :param show_sem: if true, the standard error of the mean will be shown
    :param sem_alpha: value between 0 and 1 indicating the alpha level of the sem
    :param show_sd: if true, the standard deviation of the mean will be shown
    :param sd_alpha: value between 0 and 1 indicating the alpha level of the standard deviation
    :param show_legend: if True, the legend for any other category present in the dataframe will be shown
    :param title_by: string specifying from which factor you want to show the titles in each panel. This can be: "row",
    "col", or "both"
    :param title_v_offset: float specifying the vertical offset of the title
    :param freq_vmarkers_style: style of the marker in the frequency-domain. If not passed, vertical lines are used.
    :param ylabel: Label to put in the vertical axis. If empty, the Amplitude and unit of it are used.
    :param y_unit_to: Specify the units in which the vertical axis will be displayed
    :param x_unit_to: Specify the units in which the horizontal axis will be displayed
    :param fontsize: the font size
    :param group_by: list of keys in dataframe which will be grouped. If colour_by is None, then colour is determined by
     group_by
    :param colour_by: list of keys in dataframe used to generate colours codes for the figure.
    :param show_column_labels: bool indicatint whether the columns labels should be shown
    :param color_map_label: string with the label that would be use of the colourmap. If empty, the default value will
    be the topographic_value
    :param topographic_maps_direction: direction ('v' or 'h') of topographic maps within each panel.
    :param normalize: if True, topographic maps will be normalized within normalize_by
    :param show_sensors: if True, the position of the sensors will be shown
    :param show_sensor_label: if True, the label of the sensor will be shown
    :param n_contour_levels: number of contour lines to plot
    :param min_topographic_value: if given, the colour scale will be restricted to have this value as maximum, otherwise
    it will be determined from the data.
    :param max_topographic_value: if given, the colour scale will be restricted to have this value as minimum, otherwise
    it will be determined from the data.
    :param contour_line_width: float indicating the with (in points) of the contour lines
    :param apply_function: function to be applied to each subgroup
    :param common_legend: if True, a single common legend will be used. Otherwise, each panel will contain its own
    legend
    :return:
    """
    df = copy.copy(dataframe)
    df.reset_index(inplace=True, drop=True)
    _rows_and_cols = []
    row_conditions = np.array([''])
    col_conditions = np.array([''])
    if 'domain' not in df.keys():
        df['domain'] = Domain.time
        print('domain not defined in dataframe, assuming {:} domain'.format(Domain.time))
    if rows_by is not None:
        _rows_and_cols.append(rows_by)
        row_conditions = df[rows_by].astype("category").cat.categories
        row_domains = df.iloc[df[rows_by].index]['domain'].unique()
        unique_row_domains = row_domains.size == 1
        n_rows = row_conditions.size
    else:
        df.loc[:, 'dummy_row'] = ''
        _rows_and_cols.append('dummy_row')
        row_domains = df['domain'].unique()
        unique_row_domains = row_domains.size == 1
        n_rows = 1
    idx_rows = np.arange(row_conditions.size)
    if cols_by is not None:
        _rows_and_cols.append(cols_by)
        col_conditions = df[cols_by].astype("category").cat.categories
        n_cols = col_conditions.size
    else:
        df.loc[:, 'dummy_col'] = ''
        _rows_and_cols.append('dummy_col')
        n_cols = 1
    idx_cols = np.arange(col_conditions.size)
    _x_unit_to = x_unit_to
    _y_unit_to = y_unit_to
    # set colurs from colormap
    df['dummy_color_categories'] = ''
    iterate_by = None
    new_var = pd.Series(df.loc[:, ['dummy_color_categories']].values[:, 0])
    if group_by is None and colour_by is None:
        n_colors = 1
        unique_color_factors = np.array([''])
    else:
        group_by = colour_by if group_by is None else group_by
        iterate_by = group_by
        sort_by = group_by
        if colour_by is not None:
            iterate_by = colour_by
            sort_by = list(set(colour_by).union(group_by))
        sep = '|' if len(iterate_by) > 1 else ''
        for _col_name in iterate_by:
            new_var = new_var + df[_col_name].values.astype(str) + sep
        df['dummy_color_categories'] = new_var
        for _col_name in iterate_by:
            df[_col_name] = pd.Categorical(df[_col_name])
        # sort to keep factor order in colours
        df.sort_values(by=iterate_by, inplace=True)
        vals, idx = np.unique(df['dummy_color_categories'], return_index=True)
        unsorted_unique = vals[np.argsort(idx)]

        df['dummy_color_categories'] = pd.Categorical(df['dummy_color_categories'],
                                                      categories=unsorted_unique)
        unique_color_factors = np.array(df['dummy_color_categories'].cat.categories.tolist())

    n_colors = unique_color_factors.size
    cmap = mpl.colormaps[color_map]
    _colors = cmap(np.linspace(0, 1, n_colors))
    _colors = [mpl.colors.to_hex(cmap(i)) for i in range(n_colors)]
    groups = df.groupby(_rows_and_cols, observed=False, sort=True)
    fig_out = plt.figure(constrained_layout=False)
    # set up subplots
    widths = [1.0] * n_cols
    heights = [1.0] * n_rows
    heights.insert(0, 0.01)  # this is for the top labels
    widths.append(0.01)  # this is for the right labels
    gs = fig_out.add_gridspec(ncols=n_cols + 1, nrows=n_rows + 1,
                              width_ratios=widths,
                              height_ratios=heights,
                              )

    all_ax_row_label = []
    all_ax_col_label = []
    # attach all axis to figure to ensure the are all created before calling them
    [plt.subplot(gs[_idx_row, _idx_col]) for _idx_row in range(gs.get_geometry()[0]) for _idx_col in
     range(gs.get_geometry()[1])]
    inset_axes = []

    if commom_topographic_value and max_topographic_value is None:
        _df = df.dropna(axis=0)
        max_topographic_value = _df.potentials.apply(lambda x: np.max(
            apply_function(x, **fun_args))).max()
    if commom_topographic_value and min_topographic_value is None:
        min_topographic_value = _df.potentials.apply(lambda x: np.min(
            apply_function(x, **fun_args))).min()
    has_colormap = False
    for _id, ((_current_row_group, _current_col_group), _group) in enumerate(groups):
        _current_group = _group
        if sort_by is not None:
            _current_group.sort_values(by=sort_by, inplace=True)
        _idx_row = idx_rows[_current_row_group == row_conditions].squeeze() + 1
        _idx_col = idx_cols[_current_col_group == col_conditions].squeeze()
        ax = plt.subplot(gs[_idx_row, _idx_col])
        _last_col_ax = plt.subplot(gs[_idx_row, n_cols])
        _columns_legends = []
        for _col in _current_group:
            _filter = _current_group[_col].apply(lambda x:
                                                 not (isinstance(x, str) or
                                                      isinstance(x, float) or
                                                      isinstance(x, int)))
            if np.any(_filter.to_numpy()):
                continue
            if colour_by is not None:
                if _col in colour_by:
                    _columns_legends.append(_col)
            else:
                if group_by is not None:
                    if _col in group_by:
                        _columns_legends.append(_col)
                elif len(_current_group[_col].unique()) > 1:
                    _columns_legends.append(_col)
        for _i, (_, _row) in enumerate(_current_group.iterrows()):
            # get color index
            _color_index, = np.where(_row['dummy_color_categories'] == unique_color_factors)
            _color_index = int(_color_index)
            _row_summary = add_waveforms_to_df(
                _row=_row,
                _y_unit_to=y_unit_to,
                _x_unit_to=x_unit_to,
                sub_average_time_buffer_size=sub_average_time_buffer_size
            )
            set_axis(ax=ax,
                     _row=_row_summary,
                     _columns_legends=_columns_legends,
                     _current_row_group=_current_row_group,
                     _current_col_group=_current_col_group,
                     _idx_row=_idx_row,
                     time_xlim=time_xlim,
                     time_ylim=time_ylim,
                     freq_xlim=freq_xlim,
                     freq_ylim=freq_ylim,
                     time_vmarkers=time_vmarkers,
                     freq_vmarkers=freq_vmarkers,
                     freq_vmarkers_style=freq_vmarkers_style,
                     show_individual_waveforms=show_individual_waveforms,
                     individual_waveforms_alpha=individual_waveforms_alpha,
                     show_mean=show_mean,
                     show_sem=show_sem,
                     sem_alpha=sem_alpha,
                     show_sd=show_sd,
                     sd_alpha=sd_alpha,
                     show_legend=show_legend,
                     title_by=title_by,
                     title_v_offset=title_v_offset,
                     fontsize=fontsize,
                     color=_colors[_color_index],
                     )
        # all_ax_row_label.append(_last_col_ax)
        # set right labels
        _last_col_ax.set_xticks([])
        _last_col_ax.set_yticks([])
        _last_col_ax.axis('off')
        if not has_twin(_last_col_ax):
            ax_row_label = _last_col_ax.twinx()
            ax_row_label.autoscale(enable=False)
            ax_row_label.set_ylabel(_current_row_group,
                                    size=fontsize,
                                    rotation=-90,
                                    fontweight='bold')
            ax_row_label.set_yticklabels([])
            ax_row_label.tick_params(
                axis='y',  # changes apply to the y-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelright=True,
                labelsize=fontsize)
            ax_row_label.spines['left'].set_visible(False)
            ax_row_label.spines['right'].set_visible(False)
            ax_row_label.spines['top'].set_visible(False)
            ax_row_label.spines['bottom'].set_visible(False)
            ax_row_label.set_xticks([])
            ax_row_label.set_yticks([])
            all_ax_row_label.append(ax_row_label)
        # set top labels
        _first_row_ax = plt.subplot(gs[0, _idx_col])
        _first_row_ax.set_xticks([])
        _first_row_ax.set_yticks([])
        _first_row_ax.axis('off')
        if not has_twin(_first_row_ax):
            ax_col_label = _first_row_ax.twiny()
            ax_col_label.autoscale(enable=False)
            if show_column_labels:
                ax_col_label.set_xlabel(_current_col_group,
                                        size=fontsize,
                                        fontweight='bold')
            ax_col_label.set_xticklabels([])
            ax_col_label.tick_params(
                axis='x',  # changes apply to the x-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=False,  # ticks along the top edge are off
                labelleft=True,
                labelsize=fontsize)
            ax_col_label.spines['left'].set_visible(False)
            ax_col_label.spines['right'].set_visible(False)
            ax_col_label.spines['top'].set_visible(False)
            ax_col_label.spines['bottom'].set_visible(False)
            ax_col_label.set_xticks([])
            ax_col_label.set_yticks([])
            all_ax_col_label.append(ax_col_label)

        # plot topographic map
        _current_group = _current_group.dropna(axis=0)
        if _current_group.shape[0] > 0:
            _current_group = _current_group.drop(_rows_and_cols,
                                                 axis=1)
            show_colormap = not (commom_topographic_value and has_colormap)
            _inset_axes = inset_topographic_map(dataframe=_current_group,
                                                ax=ax,
                                                insets_position=insets_position,
                                                direction=topographic_maps_direction,
                                                vmin=min_topographic_value,
                                                vmax=max_topographic_value,
                                                contour_line_width=contour_line_width,
                                                n_contour_levels=n_contour_levels,
                                                scale=inset_scale,
                                                fontsize=fontsize,
                                                show_colormap=show_colormap,
                                                colormap_size=colormap_size,
                                                group_label=group_by
                                                )
            inset_axes = inset_axes + _inset_axes
            has_colormap = True

            if _x_unit_to is None:
                _x_unit_to = _row_summary['x_unit']
            if _y_unit_to is None:
                _y_unit_to = _row_summary['y_unit']
    all_axes = fig_out.get_axes()
    all_axes = list(set(all_axes) - set(inset_axes))
    axis_with_data = [ax for ax in all_axes if
                      ax not in all_ax_row_label and
                      ax not in all_ax_col_label and
                      ax.has_data()]
    all_axis_with_data = np.all(axis_with_data)
    # check if there is a common legend. If so, then we keep only one across all axes
    _unique_labels = []
    _unique_handles = []
    for _ax in axis_with_data:
        _handles, _labels = _ax.get_legend_handles_labels()
        if len(_handles) > 0 and len(_labels) > 0:
            _unique_handles = _unique_handles + [_handles]
            _unique_labels = _unique_labels + [_labels]
    # common_labels = True
    # for _idx_ul in range(len(_unique_labels) - 1):
    #     diff = set(_unique_labels[_idx_ul]) - set(_unique_labels[_idx_ul + 1])
    #     if len(diff) > 0:
    #         common_labels = False
    #         break
    if common_legend and len(_unique_handles) > 0:
        for _ax in axis_with_data:
            if _ax.legend_ is not None:
                _ax.get_legend().remove()
        legend_elements = [
            Line2D([0],
                   [0],
                   color=_color,
                   lw=1,
                   label=_label) for _color, _label in zip(_colors,
                                                           unique_color_factors)]
        fig_out.legend(handles=legend_elements,
                       loc=legend_location,
                       fontsize=fontsize,
                       frameon=False,
                       ncol=len(legend_elements),
                       handlelength=1
                       )

    if ylabel is None:
        if _y_unit_to is not None:
            fig_out.supylabel('Amplitude [{:}]'.format(_y_unit_to), size=fontsize)
    else:
        fig_out.supylabel(ylabel, size=fontsize)
    if unique_row_domains and all_axis_with_data and np.all(row_domains == 'time') and n_cols > 1:
        fig_out.supxlabel('Time [{:}]'.format(_x_unit_to), size=fontsize)
    if unique_row_domains and all_axis_with_data and np.all(row_domains == 'frequency') and n_cols > 1:
        fig_out.supxlabel('Frequency [{:}]'.format(_x_unit_to), size=fontsize)

    unique_x_lim_per_row = []
    for _i_col in range(gs.get_geometry()[1] - 1):
        all_axis = [plt.subplot(gs[_i_row, _i_col]) for _i_row in range(1, gs.get_geometry()[0])]
        lims = []
        for _ax in all_axis:
            lims.append(np.array(_ax.get_xlim()))
        unique_x_lim_per_row.append(np.unique(np.array(lims), axis=0).shape[0] == 1)
    unique_x_lim = np.all(unique_x_lim_per_row)

    unique_y_lim_per_col = []
    for _i_row in range(1, gs.get_geometry()[0]):
        all_axis = [plt.subplot(gs[_i_row, _i_col]) for _i_col in range(gs.get_geometry()[1] - 1)]
        lims = []
        for _ax in all_axis:
            lims.append(np.array(_ax.get_ylim()))
        unique_y_lim_per_col.append(np.unique(np.array(lims), axis=0).shape[0] == 1)
    unique_y_lim = np.all(unique_y_lim_per_col)
    for ax in all_axes:
        if not ax.has_data() and ax not in all_ax_row_label and ax not in all_ax_col_label:
            ax.set_visible(False)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.set_xlabel('')
            continue
        ax.spines['top'].set_visible(False)
        if ax not in all_ax_col_label:
            if unique_row_domains and all_axis_with_data:
                subplotspec = ax.get_subplotspec()
                if subplotspec is not None and not subplotspec.is_last_row() and ax.has_data():
                    ax.set_xticklabels([])
                    ax.set_xlabel('')
                if n_cols > 1:
                    ax.set_xlabel('')
            elif unique_x_lim:
                subplotspec = ax.get_subplotspec()
                if subplotspec is not None and not subplotspec.is_last_row() and ax.has_data():
                    ax.set_xticklabels([])
                    ax.set_xlabel('')

            if unique_y_lim:
                subplotspec = ax.get_subplotspec()
                if subplotspec is not None and not subplotspec.is_first_col() and ax.has_data():
                    ax.set_yticklabels([])
                    ax.set_ylabel('')

        # ax.spines['left'].set_visible(True)
        ax.spines['right'].set_visible(False)
        ax.tick_params(labelsize=fontsize)

    inch = 2.54
    fig_out.set_size_inches(12.0 / inch, 2.25 * len(row_conditions) / inch)
    fig_out.subplots_adjust(top=0.98, bottom=0.08, hspace=0.0, left=0.15, right=0.95)
    # fig_out.tight_layout()
    return fig_out


def inset_topographic_map(dataframe: pd.DataFrame | None = None,
                          ax: object = None,
                          insets_position: type(list[float, float, float, float]) = None,
                          apply_function: object | None = np.mean,
                          fun_args: {} = {'axis': 2},
                          direction: str = 'v',
                          vmin: float | None = None,
                          vmax: float | None = None,
                          contour_line_width: float = 0.1,
                          n_contour_levels: int = 35,
                          fontsize: float = 8,
                          scale: float = 1,
                          show_colormap: bool = True,
                          colormap_size: float = 0.1,
                          group_label: List[str] | None = None,
                          ):
    df = dataframe.copy().reset_index(drop=True)
    if 'dummy_color_categories' in df.keys():
        df = df.drop(['dummy_color_categories'], axis=1)
    topo_label = '/'.join(df['topographic_label'].unique())

    grouping_variables = list(set(df.keys()) - set(['x',
                                                    'y',
                                                    'x_fs',
                                                    'x_unit',
                                                    'y_unit',
                                                    'potentials',
                                                    'max_distance',
                                                    'channels']))
    current_identifier = []
    for col in grouping_variables:
        df[col] = df[col].astype('category')
        if df[col].cat.categories.size > 1:
            current_identifier = current_identifier + [col]
    if group_label is not None:
        current_identifier = [item for item in current_identifier if item in group_label]
    n_topographic_maps = df.shape[0]
    if apply_function is not None:
        df['potentials'] = df['potentials'].apply(lambda x: apply_function(x, **fun_args))

    if vmin is None:
        min_sub_groups = df[df.potentials.apply(lambda x: np.size(x) > 1)]
        min_sub = min_sub_groups.potentials.apply(
            lambda x: np.min(x)).min()
    else:
        min_sub = vmin
    if vmax is None:
        max_sub_groups = df[df.potentials.apply(lambda x: np.size(x) > 1)]
        max_sub = max_sub_groups.potentials.apply(
            lambda x: np.max(x)).max()
    else:
        max_sub = vmax

    levels = np.linspace(min_sub, max_sub, n_contour_levels)

    _position = insets_position
    if direction not in ['v', 'h']:
        print('unknown direction {:}. using "v"'.format(direction))
        direction = 'v'

    if insets_position is None:
        height = scale * 0.5 / n_topographic_maps
        width = scale * 0.5 / n_topographic_maps
        if direction == 'v':
            _position = [1 - width,
                         0.25,
                         width,
                         0.5 / n_topographic_maps]
            _colormap_position = [1 - width * 1.1,
                                  0.25,
                                  width * colormap_size,
                                  0.5
                                  ]
        else:
            _position = [0.25,
                         1 - height,
                         0.5 / n_topographic_maps,
                         height]
            _colormap_position = [0.25,
                                  1 - height * 1.1,
                                  0.5,
                                  height * colormap_size
                                  ]
    else:
        if direction == 'v':
            width = scale * _position[2]
            height = scale * _position[3] / n_topographic_maps
            _position = [_position[0], _position[1], width, height]
            _colormap_position = [_position[0] * 0.9,
                                  _position[1],
                                  width * colormap_size,
                                  height * n_topographic_maps]
        else:
            width = scale * _position[2] / n_topographic_maps
            height = scale * _position[3]
            _position = [_position[0], _position[1], width, height]
            _colormap_position = [_position[0],
                                  _position[1] * 0.975,
                                  width * n_topographic_maps,
                                  height * colormap_size]

    _position = tuple(_position)
    all_insets = []
    ax_im = None
    for _i, _row in df.iterrows():
        if direction == 'v':
            _current_position = np.array(_position) + np.array([0, height * _i, 0, 0])
        else:
            _current_position = np.array(_position) + np.array([width * _i, 0, 0, 0])
        _current_topomap = _row['potentials']
        max_distance = _row['max_distance']
        inset_ax = ax.inset_axes(_current_position,
                                 transform=ax.transAxes)

        ax_im = inset_ax.imshow(_current_topomap.T,
                                origin='lower',
                                extent=(-max_distance, max_distance, -max_distance, max_distance),
                                vmin=min_sub,
                                vmax=max_sub,
                                aspect=1.0)
        ax_im.set_cmap('nipy_spectral')

        inset_ax.contour(_current_topomap.T,
                         levels,
                         origin='lower',
                         extent=(-max_distance, max_distance, -max_distance, max_distance),
                         linewidths=contour_line_width,
                         colors='k',
                         linestyles='solid')
        inset_ax.autoscale(enable=False)
        inset_ax.set_xticks([])
        inset_ax.set_yticks([])
        inset_ax.axis('off')
        if len(current_identifier) > 0:
            label = '/'.join([str(_row[_col]) for _col in current_identifier])
            if direction == 'v':
                ax_row_label = inset_ax.twinx()
                ax_row_label.autoscale(enable=False)
                ax_row_label.set_yticklabels([])
                ax_row_label.tick_params(
                    axis='y',  # changes apply to the y-axis
                    which='both',  # both major and minor ticks are affected
                    right=False,  # ticks along the bottom edge are off
                    left=False,  # ticks along the top edge are off
                    labelright=True,
                    labelsize=fontsize)
                ax_row_label.spines['left'].set_visible(False)
                ax_row_label.spines['right'].set_visible(False)
                ax_row_label.spines['top'].set_visible(False)
                ax_row_label.spines['bottom'].set_visible(False)
                ax_row_label.set_xticks([])
                ax_row_label.set_yticks([])
                ax_row_label.set_ylabel(label,
                                        size=fontsize,
                                        rotation=-90)
                all_insets.append(ax_row_label)
            else:
                ax_row_label = inset_ax.twiny()
                ax_row_label.autoscale(enable=False)
                ax_row_label.set_yticklabels([])
                ax_row_label.tick_params(
                    axis='x',  # changes apply to the y-axis
                    which='both',  # both major and minor ticks are affected
                    top=False,  # ticks along the bottom edge are off
                    bottom=False,  # ticks along the top edge are off
                    labelbottom=True,
                    labelsize=fontsize)
                ax_row_label.spines['left'].set_visible(False)
                ax_row_label.spines['right'].set_visible(False)
                ax_row_label.spines['top'].set_visible(False)
                ax_row_label.spines['bottom'].set_visible(False)
                ax_row_label.set_xticks([])
                ax_row_label.set_yticks([])
                ax_row_label.set_xlabel(label, size=fontsize)
                all_insets.append(ax_row_label)
    if show_colormap:
        if direction == 'v':
            ax_colormap = ax.inset_axes(_colormap_position,
                                        transform=ax.transAxes)
            c_bar = ax.figure.colorbar(ax_im,
                                       cax=ax_colormap,
                                       orientation='vertical',
                                       format='%.1f')
            c_bar.ax.tick_params(
                axis='y',  # changes apply to the y-axis
                which='both',  # both major and minor ticks are affected
                right=False,  # ticks along the bottom edge are off
                left=True,  # ticks along the top edge are off
                labelleft=True,
                labelright=False,
                labelsize=fontsize,
                # size=fontsize
            )
            c_bar.set_label(topo_label, fontsize=fontsize)
        else:
            ax_colormap = ax.inset_axes(_colormap_position,
                                        transform=ax.transAxes)
            c_bar = ax.figure.colorbar(ax_im,
                                       cax=ax_colormap,
                                       orientation='horizontal',
                                       format='%.1f')
            c_bar.ax.tick_params(
                axis='x',  # changes apply to the y-axis
                which='both',  # both major and minor ticks are affected
                labelsize=fontsize,
                # size=fontsize
            )
            c_bar.set_label(topo_label, fontsize=fontsize)

        tick_locator = ticker.MaxNLocator(nbins=3)
        c_bar.locator = tick_locator
        c_bar.update_ticks()
        all_insets.append(ax_colormap)
    return all_insets


def get_waveforms_and_topographic_maps_df(
        database_path: str = None,
        channels: List[str] = None,
        group_by: List[str] = None,
        data_source_waveforms: List[str] = None,
        data_source_topographic_maps: List[str] = None,
        data_source_stat_table: str = 'f_test',
        data_source_mapping: dict = None,
        topographic_value: str = 'f',
        layout: str = 'biosemi64_2_EXT.lay',
        prepare_dataframe: object = None,
        user_query_waveforms: str | None = None,
        user_query_topographic_maps: str | None = None,
        grid_size: np.complex128 = 600j,
        filters: str | None = None
):
    """
    This function will generate a dataframe with waveforms and topographic maps that can be directly used with
    plot_time_frequency_responses_and_topographic_maps to generate figures.
    :param database_path: str pointing to the database
    :param channels: list of channels to extract waveforms
    :param group_by: list grouping keys in dataframe
    :param data_source_waveforms: list of data_source(s) (within the waveforms table in the database) to extract
    waveforms
    :param data_source_topographic_maps: list of data_source(s) (within the data_source_stat_table table in the
    database) to extract the topographic value to be shown
    :param data_source_stat_table: str pointing to the data_source_stat_table in the database us to extract the
    topographic_value
    :param topographic_value: column within the selected table (data_source_stat_table) to be shown
    :param layout: the EEG layout used to plot the topographic map
    :param prepare_dataframe: arbitrary function new_df <- function(df) used to modify the dataframes as desired
    :param user_query_waveforms: This parameter can be used to include or exclude waveforms data based on a logical
    condition, e.g. 'subject_id != "S1"'
    :param user_query_topographic_maps: This parameter can be used to include or exclude data in the topographic maps
     based on a logical condition, e.g. 'subject_id != "S1"'
    :param grid_size: resolution of the topographic maps
    :param filters: pandas query string applied after the prepare_dataframe function has finished.
    This query will be applied to both waveforms and topographic maps dataframes
    :param data_source_mapping: dictionary to map data_source_waveformes and data_source_topographic_maps
    :return: pandas dataframe with waveforms and topographic maps that can be used directly with
    plot_time_frequency_responses_and_topographic_maps to generate figures
    """

    df = sqlite_all_waveforms_to_pandas(database_path=database_path,
                                        channels=channels,
                                        user_query=user_query_waveforms
                                        )
    if prepare_dataframe is not None:
        df = prepare_dataframe(df)

    if filters is not None:
        df = df.query(filters)
    df_waveforms_grouped = group_waveforms_df_by(df,
                                                 group_by=group_by)
    df_waveforms_grouped = df_waveforms_grouped[(df_waveforms_grouped['data_source'].isin(data_source_waveforms))]
    df_topo_maps = sqlite_tables_to_pandas(database_path=database_path,
                                           tables=[data_source_stat_table],
                                           user_query=user_query_topographic_maps)[data_source_stat_table]
    if prepare_dataframe is not None:
        df_topo_maps = prepare_dataframe(df_topo_maps)

    df_topo_maps = df_topo_maps[(df_topo_maps['data_source'].isin(data_source_topographic_maps))]

    if filters is not None:
        df_topo_maps = df_topo_maps.query(filters)

    df_topo = get_all_topographic_maps(dataframe=df_topo_maps,
                                       channels_column='channel',
                                       topographic_value=topographic_value,
                                       layout=layout,
                                       group_by=group_by,
                                       grid_size=grid_size
                                       )
    df_topo_grouped = group_topographic_df_by(df_topo, group_by=group_by)
    if 'data_source' not in group_by:
        df_waveforms_grouped['data_source'] = ''
        df_topo_grouped['data_source'] = ''
    elif data_source_mapping is not None:
        wf_ds = df_waveforms_grouped['data_source'].unique()
        tp_ds = df_topo_grouped['data_source'].unique()
        for _key, _value in data_source_mapping.items():
            if _key in wf_ds:
                df_waveforms_grouped['data_source'] = df_waveforms_grouped['data_source'].replace(_key, _value)
            if _key in tp_ds:
                df_topo_grouped['data_source'] = df_topo_grouped['data_source'].replace(_key, _value)
    _group_by = group_by
    _group_by = list(set(_group_by).intersection(df_waveforms_grouped.keys()))
    _group_by = list(set(_group_by).intersection(df_topo_grouped.keys()))
    # df_out = pd.concat([df_waveforms_grouped.set_index(_group_by),
    #                     df_topo_grouped.set_index(_group_by)],
    #                    axis=1,
    #                    join='outer').reset_index()
    df_out = pd.merge(df_waveforms_grouped, df_topo_grouped,
                      left_on=_group_by,
                      right_on=_group_by,
                      how='outer')
    return df_out



import copy
import numpy as np
import astropy.units as u
import pandas as pd
from typing import List


class SingleEvent(object):
    def __init__(self,
                 code: float | None = None,
                 time_pos: u.quantity.Quantity | None = None,
                 dur: u.quantity.Quantity | None = None,
                 bad_event: bool = False,
                 description: str | None = None):
        """
        Class for a single (trigger) event, which is characterized by having a numerical identifier (code), an
        associated time point [s], and a duration [s].
        Parameters
        ----------
        code: Trigger event code (usually some number)
        time_pos: Time in seconds of trigger event
        dur: Duration of trigger event in seconds
        bad_event: indicate if this event is considered bad
        """
        self.code = code
        self.time_pos = time_pos
        self.dur = dur
        self.bad_event = bad_event
        self.description = description

    def __repr__(self):
        return (f'{self.__class__.__name__}('f'{self.code!r}, {self.time_pos!r}, {self.dur!r}')

    def __str__(self):
        return 'Single event at time {} s with code {} and duration {} s'.format(self.time_pos, self.code, self.dur)


class Events(object):
    def __init__(self, events: np.array([SingleEvent]) = None):
        """
        Defines event class
        :param events: raw events array
        """
        self._events = events
        self._mask = None

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value: float | None = None):
        self._mask = value
        print('Event mask {:} set'.format(self.mask))
        print(self.summary().to_string(index=False))

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'{self.events().shape!r}')

    def __str__(self):
        return 'Event array of size {}'.format(self.events().shape)

    def events(self, include_bad_events: bool = False):
        out = []
        if self.mask is not None:
            for _ev in self._events:
                new_event = copy.copy(_ev)
                new_event.code = float(int(new_event.code) & int(self.mask))
                out.append(new_event)
            out = np.array(out)
        else:
            if include_bad_events:
                out = self._events
            else:
                out = np.array([_ev for _ev in self._events if not _ev.bad_event])
        return out

    def get_events(self,  code: List[float] | None = None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev for _ev in self.events(include_bad_events=include_bad_events)])
        else:
            if not (isinstance(code, list) | isinstance(code, np.ndarray)):
                code = [code]
            return np.array([_ev for _ev in self.events(include_bad_events=include_bad_events) if _ev.code in code])

    def recode_events(self,
                      every: int = 1,
                      start_at: int = 0,
                      code: List[float] | None = None,
                      new_code: float | None = None,
                      include_bad_events: bool = False):
        all_events = np.array([_ev for _ev in self.events(include_bad_events=True)])
        if code is not None:
            if not (isinstance(code, list) | isinstance(code, np.ndarray)):
                code = [code]
            _events = np.array([_ev for _ev in self.events(include_bad_events=include_bad_events) if _ev.code in code])
        else:
            _events = np.array([_ev for _ev in self.events(include_bad_events=include_bad_events)])
        _idx = []
        if every < 0:
            _target_events = np.flip(_events)[start_at::-every]
        else:
            _target_events = _events[start_at::every]
        for _ev in _target_events:
            _idx.append(int(np.where(all_events == _ev)[0]))
            _ev.code = new_code
        all_events[_idx] = _target_events
        self._events = all_events

    def get_events_index(self,  code: List[float] | None = None, fs=None, include_bad_events: bool = False):
        if code is None:
            return np.array([int(np.round(_ev.time_pos * fs)) for
                             _ev in self.events(include_bad_events=include_bad_events)])
        else:
            if not (isinstance(code, list) | isinstance(code, np.ndarray)):
                code = [code]
            return np.array([int(np.round(_ev.time_pos * fs)) for
                             _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code in code])

    def get_events_time(self, code: List[float] | None = None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev.time_pos.to(u.s).value for
                             _ev in self.events(include_bad_events=include_bad_events)]) * u.s
        else:
            if not (isinstance(code, list) | isinstance(code, np.ndarray)):
                code = [code]
            return np.array([_ev.time_pos.to(u.s).value for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code in code]) * u.s

    def get_events_duration(self, code: List[float] | None = None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev.dur.to(u.s).value for _ev in self.events(include_bad_events=include_bad_events)]) * u.s
        else:
            if not (isinstance(code, list) | isinstance(code, np.ndarray)):
                code = [code]
            return np.array([_ev.dur.to(u.s).value for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code in code]) * u.s

    def get_events_code(self, code: List[float] | None = None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev.code for _ev in self.events(include_bad_events=include_bad_events)])
        else:
            if not (isinstance(code, list) | isinstance(code, np.ndarray)):
                code = [code]
            return np.array([_ev.code for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code in code])

    def set_events_status(self,
                          event_idx: np.array,
                          bad_event: bool = False):
        for _ev in self._events[event_idx]:
            _ev.bad_event = bad_event

    def get_events_object(self,  code: List[float] | None = None, include_bad_events: bool = False):
        return Events(self.get_events(code=code, include_bad_events=include_bad_events))

    def interpolate_events_between_events(self, scaling_factor=1.0, code: List[float] | None = None, events_mask=None):
        """
        This function interpolates time events between consecutive event. The time position of new events is defined
        by the scaling factor. For example, an scaling_factor of 2.0 will add a new trigger event exactly in the middle
        time between two events
        :param scaling_factor: a float that determine the number of new events that will be generated
        :param code: list of event codes to be interpolated
        :param events_mask: integer value used to masker triggers codes. This is useful to ignore triggers inputs above
        a particular value. For example, if only the first 8 trigger inputs were used (max decimal value is 255), in a
        system with 16 trigger inputs, then the masker could be set to 255 to ignore any trigger from trigger inputs 9
        to 16.
        :return: new an Events class containing the new interpolated events
        """
        self.mask = events_mask
        _times = self.get_events_time(code=code)
        _durations = self.get_events_duration(code=code)
        _codes = self.get_events_code(code=code)
        _times_intervals = np.diff(_times)
        _times_intervals = np.append(_times_intervals, np.mean(_times_intervals))
        _new_events = []
        for _ini_time, _length, _dur, _code in zip(_times, _times_intervals, _durations, _codes):
            _new_times = np.arange(_ini_time.to(u.s).value,
                                   (_ini_time + _length).to(u.s).value,
                                   (_length / scaling_factor).to(u.s).value) * u.s
            [_new_events.append(SingleEvent(time_pos=_t, dur=_dur, code=_code)) for _t in _new_times]
        return Events(events=np.array(_new_events))

    def interpolate_events_constant_rate(self,
                                         interpolation_rate: type(u.Quantity) | None = None,
                                         code: List[float] | None = None,
                                         events_mask: int | None = None):
        """
        This function interpolates time events between consecutive events at a constant rate defined
        by the interpolation_rate. For example, an interpolation_rate of 10 * u.Hz will add a new trigger every 1 / 10
        seconds.
        :param interpolation_rate: Frequency at which events should be interpolated
        :param code: list of event codes to be interpolated
        :param events_mask: integer value used to masker triggers codes. This is useful to ignore triggers inputs above
        a particular value. For example, if only the first 8 trigger inputs were used (max decimal value is 255), in a
        system with 16 trigger inputs, then the masker could be set to 255 to ignore any trigger from trigger inputs 9
        to 16.
        :return: new an Events class containing the new interpolated events
        """
        self.mask = events_mask
        _times = self.get_events_time(code=code)
        _durations = self.get_events_duration(code=code)
        _codes = self.get_events_code(code=code)
        _times_intervals = np.diff(_times)
        _times_intervals = np.append(_times_intervals, np.mean(_times_intervals))
        _new_events = []
        for _ini_time, _length, _dur, _code in zip(_times, _times_intervals, _durations, _codes):
            _new_times = np.arange(_ini_time.to(u.s).value,
                                   (_ini_time + _length).to(u.s).value,
                                   (1 / interpolation_rate).to(u.s).value) * u.s
            [_new_events.append(SingleEvent(time_pos=_t, dur=_dur, code=_code)) for _t in _new_times]
        return Events(events=np.array(_new_events))

    def summary(self):
        out = pd.DataFrame()
        for i, _code in enumerate(np.unique(self.get_events_code())):
            out = pd.concat([out,
                             pd.DataFrame.from_dict(
                                 {'event_code': [_code],
                                  'n_events': [self.get_events_code(code=_code).size]})],
                            ignore_index=True)
        return out

    def clip(self, ini_time: u.Quantity = 0 * u.s, end_time: u.Quantity = 0 * u.s):
        self._events = np.array([_ev for _ev in self._events if _ev.time_pos >= ini_time])
        self._events = np.array([_ev for _ev in self._events if _ev.time_pos <= end_time])

    def set_offset(self, time_offset: u.Quantity = 0 * u.s):
        self._events = np.array([
            SingleEvent(time_pos=_ev.time_pos - time_offset,
                        dur=_ev.dur,
                        code=_ev.code) for _ev in self._events])

    def events_to_pandas(self, code: List[float] | None = None, exclude_code: List[float] | None = None):
        _events = self.get_events(code=code)
        if exclude_code is None:
            out = pd.DataFrame([
                {'time_pos': _event.time_pos.to(u.s).value,
                 'code': _event.code,
                 'dur': _event.dur.to(u.s).value,
                 'bad_event': _event.bad_event
                 } for _event in _events if _event.code])
        else:
            if not isinstance(exclude_code, list):
                exclude_code = [exclude_code]
            out = pd.DataFrame([
                {'time_pos': _event.time_pos.to(u.s).value,
                 'code': _event.code,
                 'dur': _event.dur.to(u.s).value,
                 'bad_event': _event.bad_event
                 } for _event in _events if _event.code not in exclude_code])
        return out

    def unique_event_codes_to_pandas(self, code: List[float] | None = None, exclude_code: List[float] | None = None):
        events_df = self.events_to_pandas(code=code,
                                          exclude_code=exclude_code)
        if events_df.shape[0]:
            events_df = events_df.drop(columns=['time_pos', 'dur', 'bad_event'])
            unique_event_codes = events_df.drop_duplicates()
        else:
            unique_event_codes = events_df
        return unique_event_codes

    def group_by(self,
                 minimum_number_events_grouping: int = 1,
                 minimum_number_events_within_group: int = 1,
                 include_bad_events: bool = False,
                 start_ini_time_offset: u.Quantity = 0 * u.s,
                 end_ini_time_offset: u.Quantity = 0 * u.s,) -> List[object]:
        """
        This method will generate new Events objects based on the grouping parameters.
        Groups will be generated for each event, beased on the time interval spanning from the first to last time event
        for each event code.
        Each Events object will contain all the events with event times falling within the time interval covered by each
        event code. Only event codes with a minimum of minimum_number_events_grouping will be considered for the group
        generation, otherwise, the event code will not be included.
        For each generated Event object, only event codes with >=  minimum_number_events_within_group will be kept.

        :param minimum_number_events_grouping: minimum number of events per code to generate a new Events object
        :param minimum_number_events_within_group: minimum number of events per code within a new group. If the number
        of events for a given code fall below this value, that event code will not be included in the group.
        :param include_bad_events: whether to included or not events tagged as bad.
        :param start_ini_time_offset: additional time offset to include in the time interval (starting time)
        :param end_ini_time_offset: additional time offset to include in the time interval (ending time)
        :return: list of new Events objects
        """
        events_summary = self.summary()
        codes = events_summary[events_summary.n_events >= minimum_number_events_grouping].event_code
        events_list = []
        for code in codes:
            times = self.get_events_time(code=code, include_bad_events=include_bad_events)
            ini_time = np.min(times) + start_ini_time_offset
            end_time = np.max(times) + end_ini_time_offset
            _events = np.array([_ev for _ev in self.get_events(include_bad_events=include_bad_events) if
                                _ev.time_pos >= ini_time])
            _events = np.array([_ev for _ev in _events if _ev.time_pos <= end_time])
            _event_object = Events(events=_events)
            _event_object_summary = _event_object.summary()
            _final_codes = _event_object_summary[
                _event_object_summary.n_events > minimum_number_events_within_group].event_code
            events_list.append(Events(events=_event_object.get_events(code=_final_codes.values)))
        return events_list



import time
import tqdm
from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.storage import SaveToDatabase
from peegy.definitions.events import Events
import numpy as np
from typing import List


class EpochsByStep(InputOutputProcess):
    def __init__(self,
                 origin_input_processes: List[InputOutputProcess] | None = None,
                 at_each_epoch_block_do: PipePool | None = None,
                 epochs_step_size: int = 2,
                 max_epochs: int | None = None,
                 deep_break: bool = True,
                 event_code: float | None = None,
                 include_all_epochs: bool = False,
                 **kwargs) -> InputOutputProcess:
        """
        Process input pipeline by blocks of epochs. It assumed that the origin_input_process an EpochData class
        (epoched data).
        :param origin_input_processes: List of InputProcess holding all the data (in epochs, or events) which will be
        used for epoch iterations
        :param at_each_epoch_block_do: The pipeline that will be processed by epochs blocks
        :param epochs_step_size: indicate the step size to run the pipeline. The number of epochs will increase by this
        number at each run.
        :param deep_break: if True, epochs block will be extracted across all possible data. This means that if you
        have 100 epochs and epochs_step_size = 2, then epochs 0-1, 2-3, ..., 99-100, then 0-3, 4-7, ..., 97-100,
        ... until 0-100 will be run when deep_break is true. Otherwise, we will only extract 0-1, 0-3, ..., 0-99 epoch
        blocks (i.e., a progressive number of epochs per iteration).
        :param event_code: event code to be use in the case that something needs to be epoched
        :param include_all_epochs: if False, the number of epochs to be processed will be strictly spaced by
        epochs_step_size until max_epochs is reached.
        If include_all_epochs is set to True, epochs will be process at each epochs_step_size < max_epochs. However,
        it will also include an additional block including all epochs.
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(EpochsByStep, self).__init__(input_process=None, **kwargs)
        self.origin_input_processes = origin_input_processes
        self.at_each_epoch_block_do = at_each_epoch_block_do
        self.epochs_step_size = epochs_step_size
        self.max_epochs = max_epochs
        self.deep_break = deep_break
        self.event_code = event_code
        self.include_all_epochs = include_all_epochs
        self._origin_input_process = None
        self.last_run = None
        self._data_containers = []
        self._event_containers = []
        self._n_epochs = None

    def transform_data(self):
        # the origin input_process is copied to ensure nothing is done to the original reference process
        n_epochs = []
        if self.origin_input_processes is None:
            _process = self.at_each_epoch_block_do[list(self.at_each_epoch_block_do)[0]].input_process
            self.origin_input_processes = [_process]
            print('No origin_input_processes provided. '
                  'Using {:} as origin data source (from at_each_epoch_block_do)'.format(_process.name))
        for _process in self.origin_input_processes:
            if _process.output_node.data.ndim == 3:
                self._data_containers.append(
                    {'process': _process,
                     'original_data': _process.output_node.data.copy()})
                n_epochs.append(_process.output_node.data.shape[2])
            if _process.output_node.data.ndim == 2 and _process.output_node.events is not None:
                self._event_containers.append(
                    {'process': _process,
                     'original_events': _process.output_node.events.get_events(code=self.event_code)})
                n_epochs.append(len(_process.output_node.events.get_events(code=self.event_code)))

        if np.unique(np.array(n_epochs)).size > 1:
            print('The origin_input_processes have different number of epochs: {:}. '
                  'Using the minimum to proceed.'.format(n_epochs))

        self._n_epochs = np.min(np.array(n_epochs))
        start_time = time.time()
        self.pipeline_per_epoch_block()
        end_time = time.time()
        print('Per epoch pipeline took{:1.3} seconds'.format(end_time - start_time))

    def pipeline_per_epoch_block(self):
        pipe_pool = self.at_each_epoch_block_do
        epochs_step_size = self.epochs_step_size
        max_epochs = self.max_epochs
        include_all_epochs = self.include_all_epochs

        if max_epochs is None:
            max_epochs = self._n_epochs
        # defines the maximum number of epoch.
        n_epochs = np.arange(epochs_step_size,
                             max_epochs + epochs_step_size,
                             epochs_step_size)
        n_epochs = n_epochs[np.argwhere(n_epochs <= self._n_epochs).flatten()]
        if include_all_epochs:
            n_epochs = np.concatenate((n_epochs, [self._n_epochs]))
        n_epochs = np.unique(n_epochs)
        _local_pipe_pool = pipe_pool
        for _epochs in tqdm.tqdm(n_epochs, desc='Processing epochs', colour='blue'):
            _n_epochs_range = np.minimum(self._n_epochs, max_epochs)
            n_blocks = int(_n_epochs_range / _epochs)
            ini_b = np.arange(0, n_blocks * _epochs, _epochs).astype(int)
            end_b = np.minimum(ini_b + int(_epochs),
                               _n_epochs_range).astype(int)
            if self.deep_break:
                _epoch_groups = zip(ini_b, end_b)
            else:
                _epoch_groups = zip([ini_b[0]], [end_b[0]])

            for _block, (_ini_b, _end_b) in tqdm.tqdm(enumerate(_epoch_groups),
                                                      desc='epoch size: {:}'.format(_epochs)):
                # update origing references
                for _process in self._data_containers:
                    _process['process'].output_node.data = _process['original_data'][:, :, _ini_b: _end_b]
                    _n_epochs = _process['process'].output_node.data.shape[2]
                for _process in self._event_containers:
                    _events_to_use = _process['original_events'][_ini_b: _end_b]
                    _process['process'].output_node.events = Events(events=_events_to_use)
                    _n_epochs = _process['process'].output_node.events.get_events().size
                for _key, _value in pipe_pool.items():
                    if isinstance(_local_pipe_pool[_key], SaveToDatabase):
                        _local_pipe_pool[_key].stimuli_information['n_epochs'] = _n_epochs
                        _local_pipe_pool[_key].stimuli_information['epoch_block'] = _block
                _local_pipe_pool.run(force=True)

        # assign last run
        self.last_run = _local_pipe_pool



import astropy.units as u
import numpy as np
from peegy.definitions.events import Events, SingleEvent


def de_epoch(data: np.array(u.Quantity) = None,
             demean_edges=True,
             edge_duration: type(u.Quantity) | None = None,
             event_code: int | None = None,
             event_duration: type(u.Quantity) | None = None,
             fs: type(u.Quantity) | None = None,
             event_offset: type(u.Quantity) | None = 0 * u.s
             ):
    if demean_edges:
        if edge_duration is None or edge_duration == 0 * u.s:
            _edge_samples = 1
        else:
            _edge_samples = int(edge_duration * fs)

        for _i in range(1, data.shape[2]):
            _previous_edge = np.mean(data[-_edge_samples::, :, _i - 1], axis=0, keepdims=True)
            _current_edge = np.mean(data[0:_edge_samples, :, _i], axis=0, keepdims=True)
            data[:, :, _i] = data[:, :, _i] - (_current_edge - _previous_edge)

    output_data = np.reshape(np.transpose(data, [0, 2, 1]), [-1, data.shape[1]], order='F')
    events = []
    if event_code is None:
        event_code = 1
    if event_duration is None:
        # ensure at least 3 samples
        event_duration = 3 / fs
    for _i in range(data.shape[2]):
        events.append(SingleEvent(code=event_code,
                                  time_pos=_i * data.shape[0] / fs + event_offset,
                                  dur=event_duration))
    events = Events(events=np.array(events))
    return output_data, events
